//	ColourPalette.h - Base class for the various colour palette generators.
//  ----------------------------------------------------------------------------
//	This file is part of 'Hummingbird', an audiovisual ecosystem thing.
//	Copyright (C) 2014  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef COLOURPALETTE_H_
#define COLOURPALETTE_H_

#include "ThreeFloats.h"

///	Base class for the various colour palette generators.
/*!
	A colour palette is simply a 1D line (0->1) with colours along it. When
	one of the getColour() methods is called it will return the colour that
	matches the passed-in point in the palette. Whether that is an interpolated
	version of two static colours or something else is entirely down to the
	subclass and how it generates its palette.
 */
class ColourPalette
{
  public:
	///	Destructor.
	virtual ~ColourPalette() {};

	///	Returns a copy of this palette.
	/*!
		Used by PaletteFactory to construct new palettes.
	 */
	virtual ColourPalette *clone() const = 0;

	///	Returns the colour at point *val* in the palette.
	virtual ThreeFloats getColour(float val) const = 0;
	///	Returns a colour that can be used as an outline for the colour at point *val* in the palette.
	virtual ThreeFloats getOutlineColour(float val) const = 0;
	///	Returns a colour that will contrast with the colour at point *val* in the palette.
	virtual ThreeFloats getContrastingColour(float val) const = 0;
	///	Returns a background colour at point *val* in the palette.
	/*!
		It is assumed that the background colours will be quantitavely
		different to the foreground colours so that foreground objects don't
		disappear into the background.
	 */
	virtual ThreeFloats getBackgroundColour(float val) const = 0;
	///	Returns a cloud colour at point *val* in the palette.
	virtual ThreeFloats getCloudColour(float val) const = 0;

	///	Returns how rare this palette should be.
	/*!
		\return 0 == not at all rare; 1 == so rare it never appears.
	 */
	virtual float getRarity() const = 0;
};

#endif
