//	CubicMeshDrawer.cpp - Draws a triangle mesh with 'cubic' colour interp.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "drawing/GLESDrawer.h"
#include "CubicMeshDrawer.h"
#include "DebugHeaders.h"

//------------------------------------------------------------------------------
CubicMeshDrawer::CubicMeshDrawer():
positionLoc(0),
uvLoc(1),
colour0Loc(2),
colour1Loc(3),
colour2Loc(4),
colour3Loc(5),
isInitialised(false),
reload(true)
{

}

//------------------------------------------------------------------------------
CubicMeshDrawer::~CubicMeshDrawer()
{

}

//------------------------------------------------------------------------------
void CubicMeshDrawer::initialise(shared_ptr<Drawer>& d)
{
	drawer = dynamic_pointer_cast<GLESDrawer>(d);

	if(vertexData.size() < 20)
		vertexData.resize(20, 0.0f);

	//--------------------------------------------------------------------------
	//Load shader.
	{
		GLuint vertexObject, fragmentObject;
		GLuint shaderProgram;
		const char *vertexChars = static_cast<const char *>(vertexString);
		const char *fragmentChars = static_cast<const char *>(fragmentString);
		GLint err;
		
		//Create the shader objects.
		vertexObject = glCreateShader(GL_VERTEX_SHADER_ARB);
		fragmentObject = glCreateShader(GL_FRAGMENT_SHADER_ARB);

		//Read the shader source into the shader objects.
		glShaderSource(vertexObject, 1, static_cast<const GLchar**>(&vertexChars), NULL);
		glShaderSource(fragmentObject, 1, static_cast<const GLchar**>(&fragmentChars), NULL);

		//Compile the shaders.
		glCompileShader(vertexObject);
		glCompileShader(fragmentObject);

		//Test they both compiled okay.
		glGetShaderiv(vertexObject, GL_COMPILE_STATUS, &err);
		if(err == GL_FALSE)
			DebugStatement(L"GLESDrawer: loadShader(); Could not compile Vertex shader CubicInterp.");
		glGetShaderiv(fragmentObject, GL_COMPILE_STATUS, &err);
		if(err == GL_FALSE)
			DebugStatement(L"GLESDrawer: loadShader(); Could not compile Fragment shader CubicInterp.");

		//Create the program object.
		shaderProgram = glCreateProgram();

		if(shaderProgram)
		{
			//Attach the shaders to the program.
			glAttachShader(shaderProgram, vertexObject);
			glAttachShader(shaderProgram, fragmentObject);

			glBindAttribLocation(shaderProgram, positionLoc, "avlVertexPosition");
			glBindAttribLocation(shaderProgram, uvLoc, "uvCoords");
			glBindAttribLocation(shaderProgram, colour0Loc, "colour0");
			glBindAttribLocation(shaderProgram, colour1Loc, "colour1");
			glBindAttribLocation(shaderProgram, colour2Loc, "colour2");
			glBindAttribLocation(shaderProgram, colour3Loc, "colour3");

			//Finally, link the program.
			glLinkProgram(shaderProgram);

			//Test the linking went okay.
			glGetProgramiv(shaderProgram, GL_LINK_STATUS, &err);
			if(err == GL_FALSE)
				DebugStatement(L"GLESDrawer: loadShader(); Could not link shader CubicInterp.");

			glUseProgram(shaderProgram);

			//Setup the shader's various uniforms.
			projectionLoc = glGetUniformLocation(shaderProgram, "avlProjectionMatrix");
			matrixLoc = glGetUniformLocation(shaderProgram, "avlModelviewMatrix");

			//Output the info log.
			DebugStatement(L"Vertex shader infolog for CubicInterp.");
			printInfoLog(vertexObject, false);
			DebugStatement(L"Fragment shader infolog for CubicInterp.");
			printInfoLog(fragmentObject, false);
			DebugStatement(L"Shader program infolog for CubicInterp.");
			printInfoLog(shaderProgram, true);

			shader = shaderProgram;

			//Setup projection and model matrices.
			glViewport(0,
					   0,
					   (int)drawer->getWindowSize().x,
					   (int)drawer->getWindowSize().y);
			projectionMatrix.generateOrthoProjection(0.0f,
													 drawer->getDrawingAreaSize().x,
													 drawer->getDrawingAreaSize().y,
													 0.0f);
			glUniformMatrix4fv(projectionLoc,
							   1,
							   GL_FALSE,
							   projectionMatrix.getMatrix());
			matrix.identity();
			glUniformMatrix4fv(matrixLoc,
							   1,
							   GL_FALSE,
							   matrix.getMatrix());
		}
	}
	
	//--------------------------------------------------------------------------
	//Use our shader.
	//glUseProgram(shader);
	
	//--------------------------------------------------------------------------
	//Setup VAO.
	glGenBuffers(1, &trianglesVBO);
	glBindBuffer(GL_ARRAY_BUFFER, trianglesVBO);
	glBufferData(GL_ARRAY_BUFFER,
				 sizeof(GLfloat) * vertexData.size(),
				 &(vertexData[0]),
				 GL_STREAM_DRAW);

	glGenVertexArrays(1, &trianglesVAO);
	glBindVertexArray(trianglesVAO);

	glEnableVertexAttribArray(positionLoc);
	glEnableVertexAttribArray(uvLoc);
	glEnableVertexAttribArray(colour0Loc);
	glEnableVertexAttribArray(colour1Loc);
	glEnableVertexAttribArray(colour2Loc);
	glEnableVertexAttribArray(colour3Loc);
	glVertexAttribPointer(positionLoc,
						  2,
						  GL_FLOAT,
						  GL_FALSE,
						  sizeof(GLfloat) * 20,
						  0);
	glVertexAttribPointer(uvLoc,
						  2,
						  GL_FLOAT,
						  GL_FALSE,
						  sizeof(GLfloat) * 20,
						  (void *)(sizeof(GLfloat) * 2));
	glVertexAttribPointer(colour0Loc,
						  4,
						  GL_FLOAT,
						  GL_FALSE,
						  sizeof(GLfloat) * 20,
						  (void *)(sizeof(GLfloat) * (2+2)));
	glVertexAttribPointer(colour1Loc,
						  4,
						  GL_FLOAT,
						  GL_FALSE,
						  sizeof(GLfloat) * 20,
						  (void *)(sizeof(GLfloat) * (2+2+4)));
	glVertexAttribPointer(colour2Loc,
						  4,
						  GL_FLOAT,
						  GL_FALSE,
						  sizeof(GLfloat) * 20,
						  (void *)(sizeof(GLfloat) * (2+2+4+4)));
	glVertexAttribPointer(colour3Loc,
						  4,
						  GL_FLOAT,
						  GL_FALSE,
						  sizeof(GLfloat) * 20,
						  (void *)(sizeof(GLfloat) * (2+2+4+4+4)));

	glBindVertexArray(0);
	drawer->resetInternalCache();
	drawer->useShader(L"");

	isInitialised = true;
}

//------------------------------------------------------------------------------
void CubicMeshDrawer::draw(shared_ptr<Drawer>& d)
{
	int i, j, k;
	ThreeFloats colours[4];
	float alphas[4];
	TwoFloats uvs[4] = {TwoFloats(0.0f, 0.0f),
						TwoFloats(1.0f, 0.0f),
						TwoFloats(0.0f, 1.0f),
						TwoFloats(1.0f, 1.0f)};
	int lastVertex = mesh.size();

	//Use our shader.
	glUseProgram(shader);

	//Stream batched vertex data to GPU.
	glBindVertexArray(trianglesVAO);

	if(reload)
	{
		//Cover for the fact that mesh may not have a multiple of 4 number of
		//vertices.
		i = (mesh.size()%4);
		if(i > 0)
			lastVertex = mesh.size()-i;

		//Fill out vertexData.
		k = 0;
		for(i=0;i<lastVertex;i+=4)
		{
			colours[0] = mesh[i].colour;
			colours[1] = mesh[i+1].colour;
			colours[2] = mesh[i+2].colour;
			colours[3] = mesh[i+3].colour;

			alphas[0] = mesh[i].alpha;
			alphas[1] = mesh[i+1].alpha;
			alphas[2] = mesh[i+2].alpha;
			alphas[3] = mesh[i+3].alpha;

			for(j=0;j<8;++j)
			{
				int j2 = j;

				if(j2 < 2)
					j2 = 0;
				else if(j2 > 5)
					j2 = 3;
				else
					j2 = j-2;

				//Position.
				vertexData[k] = mesh[i+j2].position.x;
				++k;
				vertexData[k] = mesh[i+j2].position.y;
				++k;

				//UV coords.
				vertexData[k] = uvs[j2].x;
				++k;
				vertexData[k] = uvs[j2].y;
				++k;

				//Colour + alpha 0.
				vertexData[k] = colours[0].x;
				++k;
				vertexData[k] = colours[0].y;
				++k;
				vertexData[k] = colours[0].z;
				++k;
				vertexData[k] = alphas[0];
				++k;

				//Colour + alpha 1.
				vertexData[k] = colours[1].x;
				++k;
				vertexData[k] = colours[1].y;
				++k;
				vertexData[k] = colours[1].z;
				++k;
				vertexData[k] = alphas[1];
				++k;

				//Colour + alpha 2.
				vertexData[k] = colours[2].x;
				++k;
				vertexData[k] = colours[2].y;
				++k;
				vertexData[k] = colours[2].z;
				++k;
				vertexData[k] = alphas[2];
				++k;

				//Colour + alpha 3.
				vertexData[k] = colours[3].x;
				++k;
				vertexData[k] = colours[3].y;
				++k;
				vertexData[k] = colours[3].z;
				++k;
				vertexData[k] = alphas[3];
				++k;
			}
		}

		glBindBuffer(GL_ARRAY_BUFFER, trianglesVBO);
		glBufferData(GL_ARRAY_BUFFER,
					 sizeof(GLfloat) * vertexData.size(),
					 NULL,
					 GL_STREAM_DRAW);
		glBufferSubData(GL_ARRAY_BUFFER,
						0,
						sizeof(GLfloat) * vertexData.size(),
						&(vertexData[0]));
	}

	//Draw batch.
	glDrawArrays(GL_TRIANGLE_STRIP, 0, vertexData.size()/20); //20 = num floats in a single vertex.
	
	drawer->resetInternalCache();
	drawer->useShader(L"");
}

//------------------------------------------------------------------------------
void CubicMeshDrawer::setMeshSize(int size)
{
	int i, j;

	if(size < (int)mesh.size())
	{
		for(i=mesh.size();i<size;++i)
		{
			Vertex v;

			mesh.push_back(v);

			for(j=0;j<40;++j) //40 rather than 20 to account for the padding vertices.
				vertexData.push_back(0.0f);
		}
	}

	reload = true;
}

//------------------------------------------------------------------------------
void CubicMeshDrawer::setVertex(int index,
								const TwoFloats& pos,
								const ThreeFloats& col,
								float alpha)
{
	int i, j;

	//Make sure index fits inside mesh.
	if(index >= (int)mesh.size())
	{
		for(i=mesh.size();i<(index+1);++i)
		{
			Vertex v;

			mesh.push_back(v);

			for(j=0;j<40;++j)
				vertexData.push_back(0.0f);
		}
	}

	mesh[index].position = pos;
	mesh[index].colour = col;
	mesh[index].alpha = alpha;

	reload = true;
}

//------------------------------------------------------------------------------
void CubicMeshDrawer::printInfoLog(GLuint shader, bool program)
{
	char *infoLogChar;
	std::string infoLog;
	std::wstring wInfoLog;
	GLint infologLength = 0;
    GLsizei charsWritten = 0;
	
	if(program)
		glGetProgramiv(shader, GL_INFO_LOG_LENGTH, &infologLength);
	else
		glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &infologLength);

	DebugStatement(L"infologLength = " << infologLength);
	
    if(infologLength > 0)
    {
		infoLog.resize(infologLength);
		infoLogChar = const_cast<char *>(infoLog.c_str());

		if(program)
			glGetProgramInfoLog(shader, infologLength, &charsWritten, infoLogChar);
		else
			glGetShaderInfoLog(shader, infologLength, &charsWritten, infoLogChar);

		UTF8File::charToWstring(infoLog, wInfoLog);
		DebugStatement(wInfoLog);
    }
}

//------------------------------------------------------------------------------
const char CubicMeshDrawer::vertexString[] = {"\
	#version 130\n\
	\n\
	in vec4 avlVertexPosition;\n\
	in vec2 uvCoords;\n\
	\n\
	in vec4 colour0;\n\
	in vec4 colour1;\n\
	in vec4 colour2;\n\
	in vec4 colour3;\n\
	\n\
	uniform mat4x4 avlProjectionMatrix;\n\
	uniform mat4x4 avlModelviewMatrix;\n\
	\n\
	out vec2 fragUvCoords;\n\
	out vec4 fragColour0;\n\
	out vec4 fragColour1;\n\
	out vec4 fragColour2;\n\
	out vec4 fragColour3;\n\
	\n\
	void main()\n\
	{\n\
		fragUvCoords = uvCoords;\n\
		fragColour0 = colour0;\n\
		fragColour1 = colour1;\n\
		fragColour2 = colour2;\n\
		fragColour3 = colour3;\n\
		gl_Position = avlModelviewMatrix * avlVertexPosition * avlProjectionMatrix;\n\
	}\n\
"};

//------------------------------------------------------------------------------
const char CubicMeshDrawer::fragmentString[] = {"\
	#version 130\n\
	\n\
	in vec2 fragUvCoords;\n\
	in vec4 fragColour0;\n\
	in vec4 fragColour1;\n\
	in vec4 fragColour2;\n\
	in vec4 fragColour3;\n\
	\n\
	void main()\n\
	{\n\
		vec4 colx1 = mix(fragColour0, fragColour1, smoothstep(0.0, 1.0, fragUvCoords.x));\n\
		vec4 colx2 = mix(fragColour2, fragColour3, smoothstep(0.0, 1.0, fragUvCoords.x));\n\
		vec4 col = mix(colx1, colx2, smoothstep(0.0, 1.0, fragUvCoords.y));\n\
		\n\
		gl_FragColor = col;\n\
	}\n\
"};
