//	CubicMeshDrawer.h - Draws a triangle mesh with 'cubic' colour interp.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef CUBICMESHDRAWER_H_
#define CUBICMESHDRAWER_H_

#include "drawing/GLHeaders.h"
#include "drawing/GLMatrix.h"
#include "ThreeFloats.h"
#include "c0xHeaders.h"
#include "TwoFloats.h"

#include <vector>

class GLESDrawer;
class Drawer;

///	Draws a triangle mesh with 'cubic' colour interp.
class CubicMeshDrawer
{
  public:
	///	Constructor.
	CubicMeshDrawer();
	///	Destructor.
	~CubicMeshDrawer();

	///	Returns true if the drawer has already been initialised.
	bool getIsInitialised() const {return isInitialised;};
	///	Initialises the OpenGl stuff.
	void initialise(shared_ptr<Drawer>& d);

	///	Draws the mesh.
	void draw(shared_ptr<Drawer>& d);

	///	Can be used to set the size (num vertices) of the mesh.
	void setMeshSize(int size);
	///	Sets the indexed vertex's data.
	void setVertex(int index,
				   const TwoFloats& pos,
				   const ThreeFloats& col,
				   float alpha);
  private:
	///	Helper method. Prints out the infolog for a shader when DebugStatements are turned on.
	void printInfoLog(GLuint shader, bool program);

	///	Struct representing an individual vertex.
	struct Vertex
	{
		///	The vertex position.
		TwoFloats position;
		///	The vertex colour.
		ThreeFloats colour;
		///	The vertex alpha.
		float alpha;
	};

	///	The current vertex mesh.
	std::vector<Vertex> mesh;
	///	The vertex data sent to the shader.
	std::vector<GLfloat> vertexData;

	///	The vertex array object.
	GLuint trianglesVAO;
	///	The buffer array object.
	GLuint trianglesVBO;

	///	The shader's OpenGL identifier.
	GLuint shader;

	///	Location of the projectionMatrix uniform.
	GLint projectionLoc;
	///	Location of the matrix uniform.
	GLint matrixLoc;
	///	The position of the vertex position attribute.
	GLuint positionLoc;
	///	The position of the vertex uv attribute.
	GLuint uvLoc;
	///	The position of the vertex colour 0 attribute.
	GLuint colour0Loc;
	///	The position of the vertex colour 1 attribute.
	GLuint colour1Loc;
	///	The position of the vertex colour 2 attribute.
	GLuint colour2Loc;
	///	The position of the vertex colour 3 attribute.
	GLuint colour3Loc;

	///	Our projection matrix.
	GLMatrix projectionMatrix;
	///	Our modelview matrix.
	GLMatrix matrix;

	///	True if the drawer's OpenGL stuff has already been initialised.
	bool isInitialised;
	///	True if we need to reload the mesh's vertex data.
	bool reload;

	///	Our copy of the GLESDrawer.
	shared_ptr<GLESDrawer> drawer;

	///	The vertex shader as a string.
	static const char vertexString[];
	///	The fragment shader as a string.
	static const char fragmentString[];
};

#endif
