//	GamePlayState.h - The main gameplay state.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2014  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef GAMEPLAYSTATE_H_
#define GAMEPLAYSTATE_H_

#include "drawing/TextLayout.h"
#include "ParticleGenerator.h"
#include "CubicMeshDrawer.h"
#include "RainbowPalette.h"
#include "GameState.h"
#include "StateBase.h"
#include "Tweener.h"
#include "Player.h"

#include <vector>
#include <map>

class GenericFileSounder;
class StarsGlobalData;
class ExplodeFlash;
class MuzzleFlash;
class SmokePuff;
class Explosion;
class Bullet;
class Wall2;
class Wall;

///	The game's title screen.
class GamePlayState : public GameState,
					  public StateBase
{
  public:
	///	Constructor.
	GamePlayState(Application& app, SounderPtr& sound);
	///	Destructor.
	~GamePlayState();

	///	Called frequently to update the state.
	void update();
	///	Called to draw the state.
	void draw(DrawerPtr& d, float timestepInterp);

	///	Called when a key down event is received.
	void keyDown(const std::wstring& action,
				 const std::wstring& keyPressed,
				 int keyModifier);
	///	Called when a key up event is received.
	void keyUp(const std::wstring& action,
			   const std::wstring& keyPressed,
			   int keyModifier);

	///	Called when a mouse move event is received.
	void mouseMove(const TwoFloats& pos);

	void joyMove(const std::wstring& action,
				 float value,
				 int joystick,
				 int axis);

	///	Returns true if the passed-in space is free of other Players.
	/*!
		If there is another player in the space, this will attempt to push them
		off. It will only return false if the other player had nowhere to move
		to.
	 */
	bool getSpaceIsFree(const TwoFloats& pos,
						const TwoFloats& dir,
						int& recursion);
	///	Returns true if there's a wall in the way of the passed-in direction.
	bool getWall(const TwoFloats& pos, int dir) const;

	///	Adds a bullet to the game.
	void addBullet(int dir, Player *player);
	///	Adds an explosiotn to the game.
	void addExplosion(Player *player);
	///	Adds a muzzle flash to the game.
	void addMuzzleFlash(const TwoFloats& pos,
						const ThreeFloats& col,
						int dir);
	///	Adds an explosion flash to the game.
	void addExplodeFlash(const TwoFloats& pos,
						 const ThreeFloats& col);
	///	Adds a collection of smoke puffs to the game.
	void addSmokePuffs(const TwoFloats& pos);
	///	Adds a scale up icon to the game.
	void addScaleUpIcon(const std::wstring& im, const TwoFloats& pos) {};

	///	Called by an AI player when it's made its choice in the timed game.
	void aiChoiceMade(Player *player) {};

	///	Returns the sounder.
	shared_ptr<GenericFileSounder>& getSounder() {return sounder;};
	///	Returns true if the capslock is down.
	bool getCapsLock() const {return capsLock;};

	///	Returns the name of this state.
	const std::wstring getName() const {return L"GamePlayState";};
	///	This state's creator.
	static GameStateCreator<GamePlayState> creator;
  private:
	///	Helper method. Fills out armLayouts.
	void setupArms();

	///	Helper method. Handles collisions between explosions and players/bullets.
	void collideExplosions();
	///	Helper method. Handles collisions between bullets and players.
	void collideBullets();

	///	Some constants.
	enum
	{
		BackSizeX = 5,
		BackSizeY = 4,

		GridSizeX = 16,
		GridSizeY = 9
	};

	///	Top-left offset where the grid starts.
	TwoFloats gridOffset;

	///	Used to draw the background.
	CubicMeshDrawer backgroundDrawer;
	///	The grid of colours which makes up the background.
	ThreeFloats backGrid[BackSizeY][BackSizeX];
	///	Used to tween the background colours.
	std::vector<shared_ptr<Tweener<ThreeFloats> > > backTweeners;
	///	The colour palette we use to pick colours.
	RainbowPalette colourPalette;

	///	The active players.
	std::map<std::wstring, shared_ptr<Player> > players;
	///	The active bullets.
	std::vector<shared_ptr<Bullet> > bullets;
	///	The active explosions.
	std::vector<shared_ptr<Explosion> > explosions;
	///	The active muzzle flashes.
	std::vector<shared_ptr<MuzzleFlash> > muzzleFlashes;
	///	The active explosion flashes.
	std::vector<shared_ptr<ExplodeFlash> > explodeFlashes;
	///	The active smoke puffs.
	std::vector<shared_ptr<SmokePuff> > smokePuffs;
	///	The active walls.
	std::vector<shared_ptr<Wall> > walls;
	///	The active walls.
	std::vector<shared_ptr<Wall2> > walls2;

	///	The current tick value used to determine when to switch to a new action.
	float actionTick;
	///	Previous mouse position (used to increase actionTick).
	TwoFloats mousePos;

	///	Timer used during the 'join game' mode.
	int joinTimer;
	///	Timer used during the 'join game' mode.
	float joinTimerFloat;
	///	Used by joinTimer.
	uint32_t joinMillis;
	///	True if the game is in the 'join game' mode.
	bool joinMode;
	///	Used to avoid accidental triggering when a player releases their key after the countdown.
	std::vector<std::wstring> joinKeys;
	///	TextLayout for the explanation text.
	TextLayout explainLayout;
	///	Used to fade out the explanation text.
	float explainFade;
	///	Used to fade out the explanation text.
	LogTweener<float> explainTween;

	///	Used to add walls.
	int wallTimer;
	///	The current top-left position of the available playfield.
	TwoFloats wallPos;
	///	The current walls waiting to be added.
	std::vector<TwoFloats> waitingWalls;
	///	Used to gradually add the waiting walls.
	int waitingWallsTimer;
	///	Used to reduce the time between walls being added for every loop.
	int wallsWait;

	///	The winning player.
	shared_ptr<Player> winner;
	///	The game over text.
	std::wstring gameOverText;
	///	Used to stop accepting input for a while after a game's finished.
	int gameOverCount;

	///	Number of AI players to add.
	int numAI;

	///	True if we're in hold mode (actions paused on keyDown, triggered on keyUp).
	bool holdMode;

	///	True if caps lock is held down.
	bool capsLock;

	///	The particle generator used to generate particles behind the players.
	ParticleGenerator backParticles;
	///	The particle generator used to generate particles in front of the players.
	ParticleGenerator frontParticles;

	///	Used to handle screen shake.
	float shakeAmp;
	///	Used to handle screen shake.
	float shakeIndex;

	///	Vector holding available positions for player starts.
	std::vector<TwoFloats> startPositions;
	///	The available arm layouts.
	//std::vector<std::vector<std::wstring> > armLayouts;
	///	The available player images.
	std::vector<std::wstring> playerImages;

	///	Used by the tutorial screen.
	int tutAction; //0 == Move; 1 == Shoot; 2 == Double-move; 3 == Explode
	///	Used by the tutorial screen.
	shared_ptr<Player> tutPlayer;
	///	Used by the tutorial screen.
	int tutTimer;
	///	Used by the tutorial screen.
	int tutCount; //We do 3 actions of a particular type before moving on to the next one.

	///	Our copy of the sounder.
	shared_ptr<GenericFileSounder> sounder;
	///	Our copy of the global data.
	shared_ptr<StarsGlobalData> globalData;
};

#endif
