//	Player.cpp - Class representing a player.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2014  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "sound/GenericFileSounder.h"
#include "RainbowPalette.h"
#include "drawing/Drawer.h"
#include "StateBase.h"
#include "Player.h"
#include "Random.h"

//------------------------------------------------------------------------------
Player::Player(const TwoFloats& pos,
			   const std::wstring& im,
			   RainbowPalette& palette):
position(pos.x, pos.y - 32.0f),
image(im),
size(0.7f, 1.3f),
sizeTween(size, 1.0f, 0.1f),
colTween(colour, colour, 0.02f),
shakeAmp(0.0f),
shakeIndex(0.0f),
action(Random::getGlobal().getRandomInt(NumActions)),
lives(2),
actionCount(Random::getGlobal().getRandomFloat()),
actionTick(1.0f/60.0f),
actionPause(false),
triggerPause(0.0f),
posTween(position, pos, 0.1f),
bopping(false),
waiting(true),
blinking(false),
blinkCount(0),
ai(false),
timed(false),
triggerCount(120 + Random::getGlobal().getRandomInt(240))
{
	float colIndex(Random::getGlobal().getRandomFloat());

	colour = palette.getColour(colIndex);
	colIndex += 0.15f;
	if(colIndex > 1.0f)
		colIndex -= 1.0f;
	backColour = 1.0f;

	//sizeTween.setIndex(1.0f);
	colTween.setIndex(1.0f);

	//posTween.setIndex(1.0f);
}

//------------------------------------------------------------------------------
void Player::update(StateBase *state)
{
	if(!posTween.getFinished())
	{
		posTween.tick();
		if(posTween.getFinished())
		{
			if(firstGo)
			{
				firstGo = false;
				posTween.setDelta(0.05f);
			}

			if(bopping)
			{
				bopping = false;
				posTween.reset(bopStart);
			}
		}
	}
	else if(ai && (lives > 0) && !waiting && !actionPause)
	{
		--triggerCount;
		if(!triggerCount)
		{
			//triggerCount = 120 + Random::getGlobal().getRandomInt(240);
			triggerCount = 30 + Random::getGlobal().getRandomInt(60);

			if(timed)
			{
				state->aiChoiceMade(this);
				//pauseAction(true);
			}
			else
				trigger(state);
		}
	}

	if(triggerPause > 0.0f)
		triggerPause -= 0.05f;

	if(!actionPause && !waiting)
	{
		actionCount += actionTick;
		if(actionCount > 1.0f)
		{
			int newAction;
			std::vector<int> availableActions;

			if(!state->getWall(position, 0))
			{
				availableActions.push_back(0);
				availableActions.push_back(4);
				availableActions.push_back(8);
			}
			if(!state->getWall(position, 1))
			{
				availableActions.push_back(1);
				availableActions.push_back(5);
				availableActions.push_back(9);
			}
			if(!state->getWall(position, 2))
			{
				availableActions.push_back(2);
				availableActions.push_back(6);
				availableActions.push_back(10);
			}
			if(!state->getWall(position, 3))
			{
				availableActions.push_back(3);
				availableActions.push_back(7);
				availableActions.push_back(11);
			}
			availableActions.push_back(12);

			newAction = availableActions[Random::getGlobal().getRandomInt(availableActions.size())];
			if(availableActions.size() > 1)
			{
				while(newAction == action)
					newAction = availableActions[Random::getGlobal().getRandomInt(availableActions.size())];
			}
			action = newAction;

			actionCount -= 1.0f;
		}
	}

	if(blinkCount > 0)
	{
		--blinkCount;
		if(!blinkCount)
			blinking = !blinking;
	}

	if(!sizeTween.getFinished())
	{
		sizeTween.tick();
		if(sizeTween.getFinished())
		{
			if(size.x != 1.0f)
				sizeTween.reset(1.0f);
		}
	}

	if(!colTween.getFinished())
		colTween.tick();

	if(shakeAmp > 0.0f)
	{
		shakeAmp -= 0.04f;

		shakeIndex += 0.8f;
		if(shakeIndex > (2.0f * NIALL_PI))
			shakeIndex -= (2.0f * NIALL_PI);
	}
}

//------------------------------------------------------------------------------
void Player::draw(shared_ptr<Drawer>& d)
{
	const std::wstring actionImages[] = {L"ActionMove0",
										 L"ActionMove1",
										 L"ActionMove2",
										 L"ActionMove3",
										 L"ActionFire0",
										 L"ActionFire1",
										 L"ActionFire2",
										 L"ActionFire3",
										 L"ActionDouble0",
										 L"ActionDouble1",
										 L"ActionDouble2",
										 L"ActionDouble3",
										 L"ActionExplode"};
	TwoFloats pos(position);

	if(shakeAmp > 0.0f)
	{
		pos.x += sinf(shakeIndex) * fakeLog(shakeAmp) * 4.0f;
	}

	if(lives > 0)
	{
		d->addImageToBatch(L"StarBackBlur",
						   pos - (size * 40.0f),
						   size,
						   backColour);
	}

	/*for(unsigned int i=0;i<arms.size();++i)
	{
		d->addImageToBatch(arms[i],
						   pos - (size * 38.0f),
						   size,
						   colour);
	}
	d->addImageToBatch(L"StarBack",
					   pos - (size * 28.0f),
					   size,
					   colour);*/
	d->addImageToBatch(image,
					   pos - (size * 40.0f),
					   size,
					   colour);

	if(lives > 0)
	{
		if(blinking)
		{
			d->addImageToBatch(L"StarEyeClosed",
							   pos - (TwoFloats(13.0f, 18.0f)) * size);
		}
		else
		{
			d->addImageToBatch(L"StarEyeOpen",
							   pos - (TwoFloats(13.0f, 18.0f)) * size);
		}
	}
	else
		d->addImageToBatch(L"StarEyeX",
						   pos - (TwoFloats(13.0f, 18.0f)) * size);
	if(lives > 1)
	{
		if(blinking)
		{
			d->addImageToBatch(L"StarEyeClosed",
							   pos + (TwoFloats(4.0f, -18.0f)) * size);
		}
		else
		{
			d->addImageToBatch(L"StarEyeOpen",
							   pos + (TwoFloats(4.0f, -18.0f)) * size);
		}
	}
	else
		d->addImageToBatch(L"StarEyeX",
							pos + (TwoFloats(4.0f, -18.0f)) * size);

	if(waiting || !posTween.getFinished() || (lives < 1) || (triggerPause > 0.0f))
	{
		d->addImageToBatch(L"ActionBack",
						   pos - (TwoFloats(14.0f, 6.0f) * size),
						   size);
	}
	else
	{
		d->addImageToBatch(actionImages[action],
						   pos - (TwoFloats(14.0f, 6.0f) * size),
						   size);
	}
}

//------------------------------------------------------------------------------
void Player::start()
{
	waiting = false;
}

//------------------------------------------------------------------------------
void Player::stop()
{
	waiting = true;
}

//------------------------------------------------------------------------------
void Player::trigger(StateBase *state)
{
	const bool capsLock(state->getCapsLock());

	if(timed || (posTween.getFinished() && (triggerPause <= 0.0f)))
	{
		switch(action)
		{
			case 0:
				{
					int recursion(0);
					TwoFloats dir(0.0f, -80.0f);

					if(state->getSpaceIsFree(position + dir,
											 dir,
											 recursion))
					{
						posTween.setDelta(0.05f);
						posTween.reset(position + dir);
						sizeTween.reset(TwoFloats(0.8f, 1.2f));

						state->getSounder()->triggerSound(capsLock ? L"MoveHi" : L"Move", 1.66f);
					}
					else if(posTween.getFinished())
					{
						bopping = true;
						bopStart = position;
						posTween.setDelta(0.1f);
						posTween.reset(position + (dir * 0.25f));
					}
				}
				break;
			case 1:
				{
					int recursion(0);
					TwoFloats dir(80.0f, 0.0f);

					if(state->getSpaceIsFree(position + dir,
											 dir,
											 recursion))
					{
						posTween.setDelta(0.05f);
						posTween.reset(position + dir);
						sizeTween.reset(TwoFloats(1.2f, 0.8f));

						state->getSounder()->triggerSound(capsLock ? L"MoveHi" : L"Move", 1.66f);
					}
					else if(posTween.getFinished())
					{
						bopping = true;
						bopStart = position;
						posTween.setDelta(0.1f);
						posTween.reset(position + (dir * 0.25f));
					}
				}
				break;
			case 2:
				{
					int recursion(0);
					TwoFloats dir(0.0f, 80.0f);

					if(state->getSpaceIsFree(position + dir,
											 dir,
											 recursion))
					{
						posTween.setDelta(0.05f);
						posTween.reset(position + dir);
						sizeTween.reset(TwoFloats(0.8f, 1.2f));

						state->getSounder()->triggerSound(capsLock ? L"MoveHi" : L"Move", 1.66f);
					}
					else if(posTween.getFinished())
					{
						bopping = true;
						bopStart = position;
						posTween.setDelta(0.1f);
						posTween.reset(position + (dir * 0.25f));
					}
				}
				break;
			case 3:
				{
					int recursion(0);
					TwoFloats dir(-80.0f, 0.0f);

					if(state->getSpaceIsFree(position + dir,
											 dir,
											 recursion))
					{
						posTween.setDelta(0.05f);
						posTween.reset(position + dir);
						sizeTween.reset(TwoFloats(1.2f, 0.8f));

						state->getSounder()->triggerSound(capsLock ? L"MoveHi" : L"Move", 1.66f);
					}
					else if(posTween.getFinished())
					{
						bopping = true;
						bopStart = position;
						posTween.setDelta(0.1f);
						posTween.reset(position + (dir * 0.25f));
					}
				}
				break;
			case 4:
				state->addBullet(0, this);
				triggerPause = 1.0f;
				
				if(posTween.getFinished())
				{
					bopping = true;
					bopStart = position;
					posTween.setDelta(0.2f);
					posTween.reset(position + (TwoFloats(0.0f, 80.0f) * 0.25f));
				}
				break;
			case 5:
				state->addBullet(1, this);
				triggerPause = 1.0f;
				
				if(posTween.getFinished())
				{
					bopping = true;
					bopStart = position;
					posTween.setDelta(0.2f);
					posTween.reset(position + (TwoFloats(-80.0f, 0.0f) * 0.25f));
				}
				break;
			case 6:
				state->addBullet(2, this);
				triggerPause = 1.0f;

				if(posTween.getFinished())
				{
					bopping = true;
					bopStart = position;
					posTween.setDelta(0.2f);
					posTween.reset(position + (TwoFloats(0.0f, -80.0f) * 0.25f));
				}
				break;
			case 7:
				state->addBullet(3, this);
				triggerPause = 1.0f;

				if(posTween.getFinished())
				{
					bopping = true;
					bopStart = position;
					posTween.setDelta(0.2f);
					posTween.reset(position + (TwoFloats(80.0f, 0.0f) * 0.25f));
				}
				break;
			case 8:
				{
					int recursion(0);
					TwoFloats dir(0.0f, -80.0f);

					if(state->getSpaceIsFree(position + dir,
											 dir,
											 recursion))
					{
						if(state->getSpaceIsFree(position + (dir * 2.0f),
												 dir,
												 recursion))
						{
							posTween.setDelta(0.05f);
							posTween.reset(position + (dir * 2.0f));
							sizeTween.reset(TwoFloats(0.7f, 1.3f));

							state->getSounder()->triggerSound(capsLock ? L"DoubleMoveHi" : L"DoubleMove", 1.66f);
						}
						else
						{
							posTween.setDelta(0.05f);
							posTween.reset(position + dir);
							sizeTween.reset(TwoFloats(0.8f, 1.2f));

							state->getSounder()->triggerSound(capsLock ? L"MoveHi" : L"Move", 1.66f);
						}
					}
					else
					{
						bopping = true;
						bopStart = position;
						posTween.setDelta(0.1f);
						posTween.reset(position + (dir * 0.25f));
					}
				}
				break;
			case 9:
				{
					int recursion(0);
					TwoFloats dir(80.0f, 0.0f);

					if(state->getSpaceIsFree(position + dir,
											 dir,
											 recursion))
					{
						if(state->getSpaceIsFree(position + (dir * 2.0f),
												 dir,
												 recursion))
						{
							posTween.setDelta(0.05f);
							posTween.reset(position + (dir * 2.0f));
							sizeTween.reset(TwoFloats(1.3f, 0.7f));

							state->getSounder()->triggerSound(capsLock ? L"DoubleMoveHi" : L"DoubleMove", 1.66f);
						}
						else
						{
							posTween.setDelta(0.05f);
							posTween.reset(position + dir);
							sizeTween.reset(TwoFloats(1.2f, 0.8f));

							state->getSounder()->triggerSound(capsLock ? L"MoveHi" : L"Move", 1.66f);
						}
					}
					else
					{
						bopping = true;
						bopStart = position;
						posTween.setDelta(0.1f);
						posTween.reset(position + (dir * 0.25f));
					}
				}
				break;
			case 10:
				{
					int recursion(0);
					TwoFloats dir(0.0f, 80.0f);

					if(state->getSpaceIsFree(position + dir,
											 dir,
											 recursion))
					{
						if(state->getSpaceIsFree(position + (dir * 2.0f),
												 dir,
												 recursion))
						{
							posTween.setDelta(0.05f);
							posTween.reset(position + (dir * 2.0f));
							sizeTween.reset(TwoFloats(0.7f, 1.3f));

							state->getSounder()->triggerSound(capsLock ? L"DoubleMoveHi" : L"DoubleMove", 1.66f);
						}
						else
						{
							posTween.setDelta(0.05f);
							posTween.reset(position + dir);
							sizeTween.reset(TwoFloats(0.8f, 1.2f));

							state->getSounder()->triggerSound(capsLock ? L"MoveHi" : L"Move", 1.66f);
						}
					}
					else
					{
						bopping = true;
						bopStart = position;
						posTween.setDelta(0.1f);
						posTween.reset(position + (dir * 0.25f));
					}
				}
				break;
			case 11:
				{
					int recursion(0);
					TwoFloats dir(-80.0f, 0.0f);

					if(state->getSpaceIsFree(position + dir,
											 dir,
											 recursion))
					{
						if(state->getSpaceIsFree(position + (dir * 2.0f),
												 dir,
												 recursion))
						{
							posTween.setDelta(0.05f);
							posTween.reset(position + (dir * 2.0f));
							sizeTween.reset(TwoFloats(1.3f, 0.7f));

							state->getSounder()->triggerSound(capsLock ? L"DoubleMoveHi" : L"DoubleMove", 1.66f);
						}
						else
						{
							posTween.setDelta(0.05f);
							posTween.reset(position + dir);
							sizeTween.reset(TwoFloats(1.2f, 0.8f));

							state->getSounder()->triggerSound(capsLock ? L"MoveHi" : L"Move", 1.66f);
						}
					}
					else
					{
						bopping = true;
						bopStart = position;
						posTween.setDelta(0.1f);
						posTween.reset(position + (dir * 0.25f));
					}
				}
				break;
			case 12:
				sizeTween.reset(1.2f);
				state->addExplosion(this);
				triggerPause = 1.0f;
				break;
		}
		blinking = true;
		blinkCount = 15;

		actionCount = 1.0f;
	}
}

//------------------------------------------------------------------------------
void Player::setAction(int val)
{
	action = val;
}

//------------------------------------------------------------------------------
void Player::setActionTick(float val)
{
	actionTick = val;
}

//------------------------------------------------------------------------------
void Player::pauseAction(bool val)
{
	actionPause = val;
}

//------------------------------------------------------------------------------
void Player::hurt()
{
	if(lives > 0)
	{
		--lives;
		if(!lives)
		{
			colTween.reset(0.75f);
			shakeAmp = 1.0f;
		}
		else
		{
			blinking = true;
			blinkCount = 15;
		}
	}
}

//------------------------------------------------------------------------------
void Player::push(const TwoFloats& newPos)
{
	if(!posTween.getFinished())
	{
		position = posTween.getEndVal();
		if(bopping)
			position = bopStart;
	}
	posTween.reset(newPos);
}

//------------------------------------------------------------------------------
void Player::setPosition(const TwoFloats& pos)
{
	posTween.setIndex(1.0f);
	position = pos;
}

//------------------------------------------------------------------------------
void Player::setAI()
{
	ai = true;
}

//------------------------------------------------------------------------------
void Player::setTimed()
{
	timed = true;
}
