//	Player.h - Class representing a player.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2014  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef PLAYER_H_
#define PLAYER_H_

#include "ThreeFloats.h"
#include "c0xHeaders.h"
#include "TwoFloats.h"
#include "Tweener.h"

#include <string>
#include <vector>

class RainbowPalette;
class StateBase;
class Drawer;

///	Class representing a player.
class Player
{
  public:
	///	Constructor.
	Player(const TwoFloats& pos,
		   const std::wstring& im,
		   RainbowPalette& palette);

	///	Updates the player.
	void update(StateBase *state);
	///	Draws the player.
	void draw(shared_ptr<Drawer>& d);

	///	Tells the player that the game has begun.
	void start();
	///	Tells the player the game's over.
	void stop();
	///	Triggers the player's current action.
	void trigger(StateBase *state);

	///	Sets the player's action.
	void setAction(int val);
	///	Sets the current actionTick value.
	void setActionTick(float val);
	///	Pauses/unpauses the player's actionCount.
	void pauseAction(bool val);

	///	Called when the player is hit by a bullet or an explosion.
	void hurt();
	///	Called to push a player onto another tile.
	void push(const TwoFloats& newPos);

	///	Sets the player's position.
	void setPosition(const TwoFloats& pos);
	///	Makes this an AI player.
	void setAI();
	///	Informs the AI that we're in a timed match.
	void setTimed();

	///	Returns the player's position.
	const TwoFloats& getPosition() const {return position;};
	///	Returns the player's colour.
	const ThreeFloats& getColour() const {return colour;};
	///	Returns true if the player's in the process of moving.
	bool getMoving() const {return !posTween.getFinished();};
	///	Returns the player's destination if they're in the process of moving.
	const TwoFloats& getDestination() const {return posTween.getEndVal();};
	///	Returns true if the player's paused.
	bool getPaused() const {return actionPause;};
	///	Returns the player's current action.
	int getAction() const {return action;};

	///	Returns the player's arms layout.
	///	const std::vector<std::wstring>& getArms() const {return arms;};
	///	Returns the player's image.
	const std::wstring& getImage() const {return image;};
	///	Returns the player's number of lives.
	int getNumLives() const {return lives;};
  private:
	///	Constants.
	enum
	{
		NumActions = 13
	};

	///	The player's position.
	TwoFloats position;
	///	The player's colour.
	ThreeFloats colour;
	///	The player's back colour.
	ThreeFloats backColour;
	///	The player's image.
	std::wstring image;

	///	The player's size.
	TwoFloats size;
	///	Used to tween the player's size.
	SmoothstepTweener<TwoFloats> sizeTween;

	///	Used to tween the player's colour.
	SmoothstepTweener<ThreeFloats> colTween;
	///	Used to shake the player when they die.
	float shakeAmp;
	///	Used to shake the player when they die.
	float shakeIndex;

	///	The player's current action.
	int action;
	///	The player's number of lives.
	int lives;

	///	Used to determine when we switch to a new action.
	float actionCount;
	///	Used to determine when we switch to a new action.
	float actionTick;
	///	Used to pause actionCount while the player's holding down their key.
	bool actionPause;

	///	Used to prevent new actions for a brief period after firing/exploding.
	float triggerPause;

	///	Used to tween the player's position.
	SmoothstepTweener<TwoFloats> posTween;
	///	True if we're doing a 'bop' move.
	bool bopping;
	///	For when we're bopping.
	TwoFloats bopStart;

	///	True if we're in the 'join game' mode.
	bool waiting;
	///	True if the player is blinking.
	bool blinking;
	///	Used to toggle blinking.
	int blinkCount;

	///	True if this is an AI player.
	bool ai;
	///	True if we're in a timed match.
	bool timed;
	///	Used by the AI player.
	int triggerCount;

	///	Used to reset the posTween delta after the player's jumped in.
	bool firstGo;
};

#endif
