//	SDL2.0KeyManager.cpp - SDL2.0-specific KeyManager.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "SDL2.0Backend/SDL2.0KeyManager.h"

#include <sstream>

using std::wstringstream;
using std::wstring;
using std::vector;

//------------------------------------------------------------------------------
SDL2_0KeyManager::SDL2_0KeyManager()
{

}

//------------------------------------------------------------------------------
SDL2_0KeyManager::~SDL2_0KeyManager()
{

}

//------------------------------------------------------------------------------
void SDL2_0KeyManager::getActions(SDL_Keysym& keysym,
								  vector<std::wstring>& actions) const
{
	wstring keyString;

	//First prefix the string with any modifiers.
	if(keysym.mod&KMOD_CTRL)
		keyString = L"ctrl-";
	if(keysym.mod&KMOD_SHIFT)
		keyString = L"shift-";
	if(keysym.mod&KMOD_ALT)
		keyString = L"alt-";
	if(keysym.mod&KMOD_GUI)
		keyString = L"meta-";

	//Now write the actual key identifier.
	keyString.append(getKeyString(keysym.sym));

	KeyManager::getActions(keyString, actions);
}

//------------------------------------------------------------------------------
void SDL2_0KeyManager::getActions(SDL_ControllerAxisEvent& event,
								  std::vector<std::wstring>& actions) const
{
	wstringstream keyString;

	keyString << L"joy" << event.which;

	switch(event.axis)
	{
		case SDL_CONTROLLER_AXIS_LEFTX:
			if(event.value < 128) //128 so there's a bit of a dead zone.
				keyString << L"left";
			else if(event.value > 128)
				keyString << L"right";
			else
				keyString << L"centre";
			break;
		case SDL_CONTROLLER_AXIS_LEFTY:
			if(event.value < 128)
				keyString << L"up";
			else if(event.value > 128)
				keyString << L"down";
			else
				keyString << L"centre";
			break;
		case SDL_CONTROLLER_AXIS_RIGHTX:
			if(event.value < 128) //128 so there's a bit of a dead zone.
				keyString << L"left2";
			else if(event.value > 128)
				keyString << L"right2";
			else
				keyString << L"centre2";
			break;
		case SDL_CONTROLLER_AXIS_RIGHTY:
			if(event.value < 128)
				keyString << L"up2";
			else if(event.value > 128)
				keyString << L"down2";
			else
				keyString << L"centre2";
			break;
		case SDL_CONTROLLER_AXIS_TRIGGERLEFT:
			keyString << L"lefttrigger";
			break;
		case SDL_CONTROLLER_AXIS_TRIGGERRIGHT:
			keyString << L"righttrigger";
			break;
	}

	KeyManager::getActions(keyString.str(), actions);
}

//------------------------------------------------------------------------------
void SDL2_0KeyManager::getActions(SDL_ControllerButtonEvent& event,
								  std::vector<std::wstring>& actions) const
{
	wstringstream keyString;

	keyString << L"j" << (event.which + 1);

	switch(event.button)
	{
		case SDL_CONTROLLER_BUTTON_A:
			keyString << L"a";
			break;
		case SDL_CONTROLLER_BUTTON_B:
			keyString << L"b";
			break;
		case SDL_CONTROLLER_BUTTON_X:
			keyString << L"x";
			break;
		case SDL_CONTROLLER_BUTTON_Y:
			keyString << L"y";
			break;
		case SDL_CONTROLLER_BUTTON_BACK:
			keyString << L"bk";
			break;
		case SDL_CONTROLLER_BUTTON_GUIDE:
			keyString << L"gd";
			break;
		case SDL_CONTROLLER_BUTTON_START:
			keyString << L"st";
			break;
		case SDL_CONTROLLER_BUTTON_LEFTSTICK:
			keyString << L"lst";
			break;
		case SDL_CONTROLLER_BUTTON_RIGHTSTICK:
			keyString << L"rst";
			break;
		case SDL_CONTROLLER_BUTTON_LEFTSHOULDER:
			keyString << L"ls";
			break;
		case SDL_CONTROLLER_BUTTON_RIGHTSHOULDER:
			keyString << L"rs";
			break;
		case SDL_CONTROLLER_BUTTON_DPAD_UP:
			keyString << L"up";
			break;
		case SDL_CONTROLLER_BUTTON_DPAD_DOWN:
			keyString << L"dn";
			break;
		case SDL_CONTROLLER_BUTTON_DPAD_LEFT:
			keyString << L"lt";
			break;
		case SDL_CONTROLLER_BUTTON_DPAD_RIGHT:
			keyString << L"rt";
			break;
	}

	KeyManager::getActions(keyString.str(), actions);
}

//------------------------------------------------------------------------------
void SDL2_0KeyManager::getActions(SDL_JoyAxisEvent& event,
								  vector<std::wstring>& actions) const
{
	wstringstream keyString;

	keyString << L"Joy" << event.which;

	//x-axis.
	if(event.axis == 0)
	{
		if(event.value < 128) //128 so there's a bit of a dead zone.
			keyString << L"left";
		else if(event.value > 128)
			keyString << L"right";
		else
			keyString << L"centre";
	}
	//y-axis.
	else if(event.axis == 1)
	{
		if(event.value < 128)
			keyString << L"up";
		else if(event.value > 128)
			keyString << L"down";
		else
			keyString << L"centre";
	}
	//x-axis 2.
#ifdef __APPLE__
    else if(event.axis == 3)
#else
	else if(event.axis == 2)
#endif
	{
		if(event.value < 128)
			keyString << L"left2";
		else if(event.value > 128)
			keyString << L"right2";
		else
			keyString << L"centre2";
	}
	//y-axis 2.
#ifdef __APPLE__
    else if(event.axis == 4)
#else
    else if(event.axis == 3)
#endif
	{
		if(event.value < 128)
			keyString << L"up2";
		else if(event.value > 128)
			keyString << L"down2";
		else
			keyString << L"centre2";
	}
	//Triggers.
#ifdef __APPLE__
    else if(event.axis == 2)
#else
        else if(event.axis == 4)
#endif
		keyString << L"lefttrigger";
	else if(event.axis == 5)
		keyString << L"righttrigger";

	KeyManager::getActions(keyString.str(), actions);
}

//------------------------------------------------------------------------------
void SDL2_0KeyManager::getActions(SDL_JoyButtonEvent& event,
								  vector<std::wstring>& actions) const
{
	wstringstream keyString;
    int butt(event.button);

	keyString << L"j" << event.which /*<< L"button"*/ << event.button;

	KeyManager::getActions(keyString.str(), actions);
}

//------------------------------------------------------------------------------
void SDL2_0KeyManager::getActions(SDL_JoyHatEvent& event,
								  vector<std::wstring>& actions) const
{
	wstringstream keyString;

	keyString << L"joy" << event.which << L"hat" << event.hat;

	if(event.value&SDL_HAT_CENTERED)
		keyString << L"centre";
	else
	{
		if(event.value&SDL_HAT_UP)
			keyString << L"up";
		else if(event.value&SDL_HAT_DOWN)
			keyString << L"down";
		if(event.value&SDL_HAT_LEFT)
			keyString << L"left";
		else if(event.value&SDL_HAT_RIGHT)
			keyString << L"right";
	}

	KeyManager::getActions(keyString.str(), actions);
}

//------------------------------------------------------------------------------
int SDL2_0KeyManager::getModVal(SDL_Keymod mod) const
{
	int retval = 0;

	if(mod&KMOD_CTRL)
		retval = 1;
	if(mod&KMOD_SHIFT)
		retval = 2;
	if(mod&KMOD_ALT)
		retval = 3;
	if(mod&KMOD_GUI)
		retval = 4;
	if(mod&KMOD_CAPS)
		retval = 5;

	return retval;
}

//------------------------------------------------------------------------------
wstring SDL2_0KeyManager::getKeyString(SDL_Keycode key) const
{
	wstring retval;

	switch(key)
	{
		case SDLK_BACKSPACE:
			retval = L"backspace";
			break;
		case SDLK_TAB:
			retval = L"tab";
			break;
		case SDLK_CLEAR:
			retval = L"clear";
			break;
		case SDLK_RETURN:
			retval = L"return";
			break;
		case SDLK_PAUSE:
			retval = L"pause";
			break;
		case SDLK_ESCAPE:
			retval = L"escape";
			break;
		case SDLK_SPACE:
			retval = L"space";
			break;
		case SDLK_EXCLAIM:
			retval = L"!";
			break;
		case SDLK_QUOTEDBL:
			retval = L"\"";
			break;
		case SDLK_HASH:
			retval = L"#";
			break;
		case SDLK_DOLLAR:
			retval = L"$";
			break;
		case SDLK_AMPERSAND:
			retval = L"&";
			break;
		case SDLK_QUOTE:
			retval = L"'";
			break;
		case SDLK_LEFTPAREN:
			retval = L"(";
			break;
		case SDLK_RIGHTPAREN:
			retval = L")";
			break;
		case SDLK_ASTERISK:
			retval = L"*";
			break;
		case SDLK_PLUS:
			retval = L"+";
			break;
		case SDLK_COMMA:
			retval = L",";
			break;
		case SDLK_MINUS:
			retval = L"-";
			break;
		case SDLK_PERIOD:
			retval = L".";
			break;
		case SDLK_SLASH:
			retval = L"/";
			break;
		case SDLK_0:
			retval = L"0";
			break;
		case SDLK_1:
			retval = L"1";
			break;
		case SDLK_2:
			retval = L"2";
			break;
		case SDLK_3:
			retval = L"3";
			break;
		case SDLK_4:
			retval = L"4";
			break;
		case SDLK_5:
			retval = L"5";
			break;
		case SDLK_6:
			retval = L"6";
			break;
		case SDLK_7:
			retval = L"7";
			break;
		case SDLK_8:
			retval = L"8";
			break;
		case SDLK_9:
			retval = L"9";
			break;
		case SDLK_COLON:
			retval = L":";
			break;
		case SDLK_SEMICOLON:
			retval = L";";
			break;
		case SDLK_LESS:
			retval = L"<";
			break;
		case SDLK_EQUALS:
			retval = L"=";
			break;
		case SDLK_GREATER:
			retval = L">";
			break;
		case SDLK_QUESTION:
			retval = L"?";
			break;
		case SDLK_AT:
			retval = L"@";
			break;
		case SDLK_LEFTBRACKET:
			retval = L"[";
			break;
		case SDLK_BACKSLASH:
			retval = L"\\";
			break;
		case SDLK_RIGHTBRACKET:
			retval = L"]";
			break;
		case SDLK_CARET:
			retval = L"^";
			break;
		case SDLK_UNDERSCORE:
			retval = L"_";
			break;
		case SDLK_BACKQUOTE:
			retval = L"`";
			break;
		case SDLK_a:
			retval = L"a";
			break;
		case SDLK_b:
			retval = L"b";
			break;
		case SDLK_c:
			retval = L"c";
			break;
		case SDLK_d:
			retval = L"d";
			break;
		case SDLK_e:
			retval = L"e";
			break;
		case SDLK_f:
			retval = L"f";
			break;
		case SDLK_g:
			retval = L"g";
			break;
		case SDLK_h:
			retval = L"h";
			break;
		case SDLK_i:
			retval = L"i";
			break;
		case SDLK_j:
			retval = L"j";
			break;
		case SDLK_k:
			retval = L"k";
			break;
		case SDLK_l:
			retval = L"l";
			break;
		case SDLK_m:
			retval = L"m";
			break;
		case SDLK_n:
			retval = L"n";
			break;
		case SDLK_o:
			retval = L"o";
			break;
		case SDLK_p:
			retval = L"p";
			break;
		case SDLK_q:
			retval = L"q";
			break;
		case SDLK_r:
			retval = L"r";
			break;
		case SDLK_s:
			retval = L"s";
			break;
		case SDLK_t:
			retval = L"t";
			break;
		case SDLK_u:
			retval = L"u";
			break;
		case SDLK_v:
			retval = L"v";
			break;
		case SDLK_w:
			retval = L"w";
			break;
		case SDLK_x:
			retval = L"x";
			break;
		case SDLK_y:
			retval = L"y";
			break;
		case SDLK_z:
			retval = L"z";
			break;
		case SDLK_DELETE:
			retval = L"delete";
			break;
		case SDLK_KP_0:
			retval = L"numpad0";
			break;
		case SDLK_KP_1:
			retval = L"numpad1";
			break;
		case SDLK_KP_2:
			retval = L"numpad2";
			break;
		case SDLK_KP_3:
			retval = L"numpad3";
			break;
		case SDLK_KP_4:
			retval = L"numpad4";
			break;
		case SDLK_KP_5:
			retval = L"numpad5";
			break;
		case SDLK_KP_6:
			retval = L"numpad6";
			break;
		case SDLK_KP_7:
			retval = L"numpad7";
			break;
		case SDLK_KP_8:
			retval = L"numpad8";
			break;
		case SDLK_KP_9:
			retval = L"numpad9";
			break;
		case SDLK_KP_PERIOD:
			retval = L"numpad.";
			break;
		case SDLK_KP_DIVIDE:
			retval = L"numpad/";
			break;
		case SDLK_KP_MULTIPLY:
			retval = L"numpad*";
			break;
		case SDLK_KP_MINUS:
			retval = L"numpad-";
			break;
		case SDLK_KP_PLUS:
			retval = L"numpad+";
			break;
		case SDLK_KP_ENTER:
			retval = L"numpadenter";
			break;
		case SDLK_KP_EQUALS:
			retval = L"numpad=";
			break;
		case SDLK_UP:
			retval = L"cursorup";
			break;
		case SDLK_DOWN:
			retval = L"cursordown";
			break;
		case SDLK_LEFT:
			retval = L"cursorleft";
			break;
		case SDLK_RIGHT:
			retval = L"cursorright";
			break;
		case SDLK_INSERT:
			retval = L"insert";
			break;
		case SDLK_HOME:
			retval = L"home";
			break;
		case SDLK_END:
			retval = L"end";
			break;
		case SDLK_PAGEUP:
			retval = L"pageup";
			break;
		case SDLK_PAGEDOWN:
			retval = L"pagedown";
			break;
		case SDLK_F1:
			retval = L"f1";
			break;
		case SDLK_F2:
			retval = L"f2";
			break;
		case SDLK_F3:
			retval = L"f3";
			break;
		case SDLK_F4:
			retval = L"f4";
			break;
		case SDLK_F5:
			retval = L"f5";
			break;
		case SDLK_F6:
			retval = L"f6";
			break;
		case SDLK_F7:
			retval = L"f7";
			break;
		case SDLK_F8:
			retval = L"f8";
			break;
		case SDLK_F9:
			retval = L"f9";
			break;
		case SDLK_F10:
			retval = L"f10";
			break;
		case SDLK_F11:
			retval = L"f11";
			break;
		case SDLK_F12:
			retval = L"f12";
			break;
		case SDLK_F13:
			retval = L"f13";
			break;
		case SDLK_F14:
			retval = L"f14";
			break;
		case SDLK_F15:
			retval = L"f15";
			break;
		case SDLK_HELP:
			retval = L"help";
			break;
		case SDLK_PRINTSCREEN:
			retval = L"printscreen";
			break;
		case SDLK_CAPSLOCK:
			retval = L"capslock";
		default: //To avoid gcc warnings.
			break;
	}

	return retval;
}

//------------------------------------------------------------------------------
std::wstring SDL2_0KeyManager::getControllerAxisText(SDL_ControllerAxisEvent& event) const
{
	wstringstream keyString;

	keyString << L"Joy" << event.which;

	switch(event.axis)
	{
		case SDL_CONTROLLER_AXIS_LEFTX:
			if(event.value < 128) //128 so there's a bit of a dead zone.
				keyString << L"left";
			else if(event.value > 128)
				keyString << L"right";
			else
				keyString << L"centre";
			break;
		case SDL_CONTROLLER_AXIS_LEFTY:
			if(event.value < 128)
				keyString << L"up";
			else if(event.value > 128)
				keyString << L"down";
			else
				keyString << L"centre";
			break;
		case SDL_CONTROLLER_AXIS_RIGHTX:
			if(event.value < 128) //128 so there's a bit of a dead zone.
				keyString << L"left2";
			else if(event.value > 128)
				keyString << L"right2";
			else
				keyString << L"centre2";
			break;
		case SDL_CONTROLLER_AXIS_RIGHTY:
			if(event.value < 128)
				keyString << L"up2";
			else if(event.value > 128)
				keyString << L"down2";
			else
				keyString << L"centre2";
			break;
		case SDL_CONTROLLER_AXIS_TRIGGERLEFT:
			keyString << L"lefttrigger";
			break;
		case SDL_CONTROLLER_AXIS_TRIGGERRIGHT:
			keyString << L"righttrigger";
			break;
	}

	return keyString.str();
}

//------------------------------------------------------------------------------
std::wstring SDL2_0KeyManager::getControllerButtonText(SDL_ControllerButtonEvent& event) const
{
	wstringstream keyString;

	keyString << L"j" << (event.which + 1);

	switch(event.button)
	{
		case SDL_CONTROLLER_BUTTON_A:
			keyString << L"a";
			break;
		case SDL_CONTROLLER_BUTTON_B:
			keyString << L"b";
			break;
		case SDL_CONTROLLER_BUTTON_X:
			keyString << L"x";
			break;
		case SDL_CONTROLLER_BUTTON_Y:
			keyString << L"y";
			break;
		case SDL_CONTROLLER_BUTTON_BACK:
			keyString << L"bk";
			break;
		case SDL_CONTROLLER_BUTTON_GUIDE:
			keyString << L"gd";
			break;
		case SDL_CONTROLLER_BUTTON_START:
			keyString << L"st";
			break;
		case SDL_CONTROLLER_BUTTON_LEFTSTICK:
			keyString << L"lst";
			break;
		case SDL_CONTROLLER_BUTTON_RIGHTSTICK:
			keyString << L"rst";
			break;
		case SDL_CONTROLLER_BUTTON_LEFTSHOULDER:
			keyString << L"ls";
			break;
		case SDL_CONTROLLER_BUTTON_RIGHTSHOULDER:
			keyString << L"rs";
			break;
		case SDL_CONTROLLER_BUTTON_DPAD_UP:
			keyString << L"up";
			break;
		case SDL_CONTROLLER_BUTTON_DPAD_DOWN:
			keyString << L"dn";
			break;
		case SDL_CONTROLLER_BUTTON_DPAD_LEFT:
			keyString << L"lt";
			break;
		case SDL_CONTROLLER_BUTTON_DPAD_RIGHT:
			keyString << L"rt";
			break;
	}

	return keyString.str();
}

//------------------------------------------------------------------------------
wstring SDL2_0KeyManager::getJoyButtonText(SDL_JoyButtonEvent& event) const
{
	wstringstream keyString;

	keyString << L"j" << (event.which + 1) << L"b" << (event.button + 1);

	return keyString.str();
}

//------------------------------------------------------------------------------
wstring SDL2_0KeyManager::getJoyHatText(SDL_JoyHatEvent& event) const
{
	wstringstream keyString;

	keyString << L"joy" << event.which << L"hat" << event.hat;

	if(event.value&SDL_HAT_CENTERED)
		keyString << L"centre";
	else
	{
		if(event.value&SDL_HAT_UP)
			keyString << L"up";
		else if(event.value&SDL_HAT_DOWN)
			keyString << L"down";
		if(event.value&SDL_HAT_LEFT)
			keyString << L"left";
		else if(event.value&SDL_HAT_RIGHT)
			keyString << L"right";
	}

	return keyString.str();
}
