//	TitleScreen.cpp - The title screen.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2014  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "sound/GenericFileSounder.h"
#include "StarsGlobalData.h"
#include "drawing/Drawer.h"
#include "TitleScreen.h"
#include "Application.h"
#include "Random.h"

//------------------------------------------------------------------------------
GameStateCreator<TitleScreen> TitleScreen::creator;

//------------------------------------------------------------------------------
TitleScreen::TitleScreen(Application& app, SounderPtr& sound):
GameState(app, sound),
timedSize(1.0f),
timedTween(timedSize, 1.0f, 0.066f),
chaosSize(1.0f),
chaosTween(chaosSize, 1.0f, 0.066f),
holdSize(1.0f),
holdTween(holdSize, 1.0f, 0.066f),
option(0),
highlight((app.getDrawingAreaSize() * 0.5f) - TwoFloats(320.0f, 0.0f), colourPalette),
mouseCol(colourPalette.getColour(Random::getGlobal().getRandomFloat())),
mouseTween(mouseCol, colourPalette.getColour(Random::getGlobal().getRandomFloat()), 0.01f),
fade(1.0f),
fadeTween(fade, 0.0f, 0.02f),
sounder(dynamic_pointer_cast<GenericFileSounder>(sound))
{
	static bool comingFromLoadingScreen(true);
	const TwoFloats& drawSize(app.getDrawingAreaSize());

	if(comingFromLoadingScreen)
	{
		comingFromLoadingScreen = false;
		sounder->playMusic(L"Music");
	}
	else
	{
		fade = 0.0f;
		fadeTween.setIndex(1.0f);
	}

	timedTween.setIndex(1.0f);
	chaosTween.setIndex(1.0f);
	holdTween.setIndex(1.0f);

	globalData = dynamic_pointer_cast<StarsGlobalData>(app.getGlobalData());

	gridOffset.x = (drawSize.x - 1280.0f) * 0.5f;
	gridOffset.y = (drawSize.y - 720.0f) * 0.5f;

	particles.setSize(0.66f);
	particles.setLife(1.5f);

	//Setup background.
	{
		int v = 0;
		
		for(int y=0;y<BackSizeY;++y)
		{
			for(int x=0;x<BackSizeX;++x)
			{
				float index(Random::getGlobal().getRandomFloat());

				//backGrid[y][x] = colourPalette.getBackgroundColour(index);

				if((x < 1) || (x > (BackSizeX-2)) || (y < 1) || (y > (BackSizeY-2)))
					backGrid[y][x] = colourPalette.getCloudColour(Random::getGlobal().getRandomFloat());
				else
					backGrid[y][x] = colourPalette.getBackgroundColour(Random::getGlobal().getRandomFloat());

				index = Random::getGlobal().getRandomFloat();
				if((x < 1) || (x > (BackSizeX-2)) || (y < 1) || (y > (BackSizeY-2)))
				{
					backTweeners.push_back(shared_ptr<Tweener<ThreeFloats> >(new Tweener<ThreeFloats>(backGrid[y][x],
																									  colourPalette.getCloudColour(index),
																									  0.005f + (Random::getGlobal().getRandomFloat() * 0.01f))));
				}
				else
				{
						backTweeners.push_back(shared_ptr<Tweener<ThreeFloats> >(new Tweener<ThreeFloats>(backGrid[y][x],
																									  colourPalette.getBackgroundColour(index),
																									  0.005f + (Random::getGlobal().getRandomFloat() * 0.01f))));
				}
			}
		}

		for(int y=0;y<(BackSizeY-1);++y)
		{
			for(int x=0;x<(BackSizeX-1);++x)
			{
				backgroundDrawer.setVertex(v,
										   TwoFloats((float)x*(getDrawingAreaSize().x/(float)(BackSizeX-1)),
													 (float)y*(getDrawingAreaSize().y/(float)(BackSizeY-1))),
										   backGrid[y][x],
										   1.0f);
				++v;
				backgroundDrawer.setVertex(v,
										   TwoFloats((float)(x+1)*(getDrawingAreaSize().x/(float)(BackSizeX-1)),
													 (float)y*(getDrawingAreaSize().y/(float)(BackSizeY-1))),
										   backGrid[y][x+1],
										   1.0f);
				++v;
				backgroundDrawer.setVertex(v,
										   TwoFloats((float)x*(getDrawingAreaSize().x/(float)(BackSizeX-1)),
													 (float)(y+1)*(getDrawingAreaSize().y/(float)(BackSizeY-1))),
										   backGrid[y+1][x],
										   1.0f);
				++v;
				backgroundDrawer.setVertex(v,
										   TwoFloats((float)(x+1)*(getDrawingAreaSize().x/(float)(BackSizeX-1)),
													 (float)(y+1)*(getDrawingAreaSize().y/(float)(BackSizeY-1))),
										   backGrid[y+1][x+1],
										   1.0f);
				++v;
			}
		}
	}
}

//------------------------------------------------------------------------------
void TitleScreen::update()
{
	const TwoFloats& drawSize(getApplication().getDrawingAreaSize());

	//Update background colours.
	for(unsigned int i=0;i<backTweeners.size();++i)
	{
		int x(i%BackSizeX);
		int y(i/BackSizeX);

		backTweeners[i]->tick();
		if(backTweeners[i]->getFinished())
		{
			if((x < 1) || (x > (BackSizeX-2)) || (y < 1) || (y > (BackSizeY-2)))
				backTweeners[i]->reset(colourPalette.getCloudColour(Random::getGlobal().getRandomFloat()));
			else
				backTweeners[i]->reset(colourPalette.getBackgroundColour(Random::getGlobal().getRandomFloat()));
		}
	}

	highlight.update();
	particles.update();

	if(!mouseTween.getFinished())
	{
		mouseTween.tick();
		if(mouseTween.getFinished())
			mouseTween.reset(colourPalette.getColour(Random::getGlobal().getRandomFloat()));
	}

	if(!timedTween.getFinished())
	{
		timedTween.tick();
		if(timedTween.getFinished() && timedSize < 1.0f)
			timedTween.reset(1.0f);
	}
	if(!chaosTween.getFinished())
	{
		chaosTween.tick();
		if(chaosTween.getFinished() && chaosSize < 1.0f)
			chaosTween.reset(1.0f);
	}
	if(!holdTween.getFinished())
	{
		holdTween.tick();
		if(holdTween.getFinished() && holdSize < 1.0f)
			holdTween.reset(1.0f);
	}

	if(!fadeTween.getFinished())
	{
		fadeTween.tick();

		if(fadeTween.getFinished() && (fade >= 1.0f))
			quit();
	}
}

//------------------------------------------------------------------------------
void TitleScreen::draw(shared_ptr<Drawer>& d, float timestepInterp)
{
	const TwoFloats& drawSize(d->getDrawingAreaSize());

	if(!backgroundDrawer.getIsInitialised())
		backgroundDrawer.initialise(d);

	//Draw the background.
	{
		int v(0);

		for(int y=0;y<BackSizeY;++y)
		{
			for(int x=0;x<BackSizeX;++x)
			{
				backgroundDrawer.setVertex(v,
											TwoFloats((float)x*(getDrawingAreaSize().x/(float)(BackSizeX-1)),
														(float)y*(getDrawingAreaSize().y/(float)(BackSizeY-1))),
											backGrid[y][x],
											1.0f);
				++v;
				backgroundDrawer.setVertex(v,
											TwoFloats((float)(x+1)*(getDrawingAreaSize().x/(float)(BackSizeX-1)),
														(float)y*(getDrawingAreaSize().y/(float)(BackSizeY-1))),
											backGrid[y][x+1],
											1.0f);
				++v;
				backgroundDrawer.setVertex(v,
											TwoFloats((float)x*(getDrawingAreaSize().x/(float)(BackSizeX-1)),
														(float)(y+1)*(getDrawingAreaSize().y/(float)(BackSizeY-1))),
											backGrid[y+1][x],
											1.0f);
				++v;
				backgroundDrawer.setVertex(v,
											TwoFloats((float)(x+1)*(getDrawingAreaSize().x/(float)(BackSizeX-1)),
														(float)(y+1)*(getDrawingAreaSize().y/(float)(BackSizeY-1))),
											backGrid[y+1][x+1],
											1.0f);
				++v;
			}
		}
	}
	backgroundDrawer.draw(d);

	//Draw the grid.
	for(int y=0;y<GridSizeY;++y)
	{
		TwoFloats pos(gridOffset);

		if(y%2)
			pos.x += 80.0f;
		pos.y += 80.0f * (float)y;

		for(int x=0;x<GridSizeX;++x)
		{
			if(!(x%2))
			{
				/*d->fillRect(pos+TwoFloats(shadowX0 * 4.0f, shadowX1 * 4.0f),
							80.0f,
							1.0f,
							0.167f);*/
				d->fillRect(pos, 80.0f, 1.0f, 0.33f);
			}

			pos.x += 80.0f;
		}
	}

	//Draw the highlight.
	highlight.drawRays(d);
	
	//Draw the title.
	d->drawText(L"...and all those lonesome stars",
				L"BigFont",
				titleLayout,
				TwoFloats(0.0f, 128.0f) + 4.0f,
				drawSize.x,
				TextLayout::Centre,
				1.0f,
				0.0f,
				0.33f);
	d->drawText(L"...and all those lonesome stars",
				L"BigFont",
				titleLayout,
				TwoFloats(0.0f, 128.0f),
				drawSize.x,
				TextLayout::Centre,
				1.0f,
				1.0f,
				1.0f);

	//Draw the expo text.
	if(globalData->getExpoMode())
	{
		d->drawText(L"expo mode",
					L"MedFont",
					expoLayout,
					TwoFloats(0.0f, 208.0f) + 4.0f,
					drawSize.x,
					TextLayout::Centre,
					1.0f,
					0.0f,
					0.6f);
		d->drawText(L"expo mode",
					L"MedFont",
					expoLayout,
					TwoFloats(0.0f, 208.0f),
					drawSize.x,
					TextLayout::Centre,
					1.0f,
					mouseCol);
	}

	//Draw the no specials text.
	if(globalData->getNoSpecials())
	{
		d->drawText(L"no specials",
					L"MedFont",
					noSpecialsLayout,
					TwoFloats(0.0f, (drawSize.y * 0.5f) + 208.0f) + 4.0f,
					drawSize.x,
					TextLayout::Centre,
					1.0f,
					0.0f,
					0.6f);
		d->drawText(L"no specials",
					L"MedFont",
					noSpecialsLayout,
					TwoFloats(0.0f, (drawSize.y * 0.5f) + 208.0f),
					drawSize.x,
					TextLayout::Centre,
					1.0f,
					mouseCol);
	}

	//Draw the timed mode.
	d->drawImage(L"ClockBack",
				 TwoFloats((drawSize.x * 0.5f) - 384.0f, (drawSize.y * 0.5f) - 64.0f) + ((1.0f - timedSize) * 0.5f * 128.0f),
				 timedSize);
	d->drawRotImage(L"ClockHand",
					TwoFloats((drawSize.x * 0.5f) - 384.0f, (drawSize.y * 0.5f) - 64.0f) + ((1.0f - timedSize) * 0.5f * 128.0f),
					(0.33f * NIALL_PI),
					64.0f,
					timedSize);
	d->drawText(L"timed",
				L"MedFont",
				timedLayout,
				TwoFloats((drawSize.x * 0.5f) - 384.0f, (drawSize.y * 0.5f) + 128.0f) + 4.0f,
				128.0f,
				TextLayout::Centre,
				1.0f,
				0.0f,
				0.33f);
	d->drawText(L"timed",
				L"MedFont",
				timedLayout,
				TwoFloats((drawSize.x * 0.5f) - 384.0f, (drawSize.y * 0.5f) + 128.0f),
				128.0f,
				TextLayout::Centre,
				1.0f,
				1.0f);

	//Draw the chaos mode.
	d->drawImage(L"TutExplode",
				 TwoFloats((drawSize.x * 0.5f) - 64.0f, (drawSize.y * 0.5f) - 64.0f) + ((1.0f - chaosSize) * 0.5f * 128.0f),
				 chaosSize);
	d->drawText(L"chaos",
				L"MedFont",
				chaosLayout,
				TwoFloats((drawSize.x * 0.5f) - 64.0f, (drawSize.y * 0.5f) + 128.0f) + 4.0f,
				128.0f,
				TextLayout::Centre,
				1.0f,
				0.0f,
				0.33f);
	d->drawText(L"chaos",
				L"MedFont",
				chaosLayout,
				TwoFloats((drawSize.x * 0.5f) - 64.0f, (drawSize.y * 0.5f) + 128.0f),
				128.0f,
				TextLayout::Centre,
				1.0f,
				1.0f);

	//Draw the chaos (hold) mode.
	d->drawImage(L"TutShoot",
				 TwoFloats((drawSize.x * 0.5f) + 256.0f, (drawSize.y * 0.5f) - 64.0f) + ((1.0f - holdSize) * 0.5f * 128.0f),
				 holdSize);
	d->drawText(L"hold",
				L"MedFont",
				chaosHoldLayout,
				TwoFloats((drawSize.x * 0.5f) + 256.0f, (drawSize.y * 0.5f) + 128.0f) + 4.0f,
				128.0f,
				TextLayout::Centre,
				1.0f,
				0.0f,
				0.33f);
	d->drawText(L"hold",
				L"MedFont",
				chaosHoldLayout,
				TwoFloats((drawSize.x * 0.5f) + 256.0f, (drawSize.y * 0.5f) + 128.0f),
				128.0f,
				TextLayout::Centre,
				1.0f,
				1.0f);

	//Draw quit.
	d->drawText(L"quit",
				L"MedFont",
				quitLayout,
				TwoFloats(0.0f, drawSize.y - 72.0f) + 4.0f,
				drawSize.x,
				TextLayout::Centre,
				1.0f,
				0.0f,
				0.33f);
	d->drawText(L"quit",
				L"MedFont",
				quitLayout,
				TwoFloats(0.0f, drawSize.y - 72.0f),
				drawSize.x,
				TextLayout::Centre,
				1.0f,
				1.0f);

	d->clearBatch();
	particles.draw(d);
	d->drawBatch();

	if(fade > 0.0f)
		d->fillRect(0.0f, drawSize, 1.0f, fade);
}

//------------------------------------------------------------------------------
void TitleScreen::keyDown(const std::wstring& action,
						  const std::wstring& keyPressed,
						  int keyModifier)
{
	const TwoFloats& drawSize(getApplication().getDrawingAreaSize());
	TwoFloats positions[] = {TwoFloats((drawSize.x * 0.5f) - 320.0f, drawSize.y * 0.5f),
							 TwoFloats(drawSize * 0.5f),
							 TwoFloats((drawSize.x * 0.5f) + 320.0f, drawSize.y * 0.5f),
							 TwoFloats(drawSize.x * 0.5f, drawSize.y - 96.0f)};

	if(action == L"Escape")
	{
		if(fadeTween.getFinished())
		{
			fadeTween.reset(1.0f);
			sounder->stopMusic();
		}
		else
			quit();
	}
	else if(keyPressed == L"e")
		globalData->setExpoMode(!globalData->getExpoMode());
	else if(keyPressed == L"m")
	{
		if(!sounder->getMusicIsPlaying())
			sounder->playMusic(L"Music");
		else
			sounder->stopMusic();
	}
	else if(keyPressed == L"s")
		globalData->setNoSpecials(!globalData->getNoSpecials());
	else if((keyPressed == L"return") || (keyPressed == L"space"))
		mouseDown(0, 0.0f);
	else if(keyPressed == L"cursorup")
	{
		if(option == 3)
		{
			option = 0;

			timedTween.reset(0.8f);
			sounder->triggerSound(L"BulletHit", 2.33f);

			highlight.setPosition(positions[option]);
		}
	}
	else if(keyPressed == L"cursordown")
	{
		if(option != 3)
		{
			option = 3;
			highlight.setPosition(positions[option]);
		}
	}
	else if(keyPressed == L"cursorleft")
	{
		switch(option)
		{
			case 1:
				option = 0;
				timedTween.reset(0.8f);
				sounder->triggerSound(L"BulletHit", 2.33f);
				break;
			case 2:
				option = 1;
				chaosTween.reset(0.8f);
				sounder->triggerSound(L"Explode", 2.0f);
				break;
			default:
				break;
		}
		highlight.setPosition(positions[option]);
	}
	else if(keyPressed == L"cursorright")
	{
		switch(option)
		{
			case 0:
				option = 1;
				chaosTween.reset(0.8f);
				sounder->triggerSound(L"Explode", 2.0f);
				break;
			case 1:
				option = 2;
				holdTween.reset(0.8f);
				sounder->triggerSound(L"Shoot", 1.66f);
				break;
			default:
				break;
		}
		highlight.setPosition(positions[option]);
	}
}

//------------------------------------------------------------------------------
void TitleScreen::mouseMove(const TwoFloats& pos)
{
	const TwoFloats& drawSize(getApplication().getDrawingAreaSize());
	TwoFloats positions[] = {TwoFloats((drawSize.x * 0.5f) - 320.0f, drawSize.y * 0.5f),
							 TwoFloats(drawSize * 0.5f),
							 TwoFloats((drawSize.x * 0.5f) + 320.0f, drawSize.y * 0.5f),
							 TwoFloats(drawSize.x * 0.5f, drawSize.y - 96.0f)};

	if(pos.y > (drawSize.y - 192.0f))
		option = 3;
	else if(pos.x < ((drawSize.x * 0.5f) - 128.0f))
	{
		if(option != 0)
		{
			timedTween.reset(0.8f);
			sounder->triggerSound(L"BulletHit", 2.33f);
		}
		option = 0;
	}
	else if(pos.x < ((drawSize.x * 0.5f) + 128.0f))
	{
		if(option != 1)
		{
			chaosTween.reset(0.8f);
			sounder->triggerSound(L"Explode", 2.0f);
		}
		option = 1;
	}
	else
	{
		if(option != 2)
		{
			holdTween.reset(0.8f);
			sounder->triggerSound(L"Shoot", 1.66f);
		}
		option = 2;
	}

	highlight.setPosition(positions[option]);

	TwoFloats gravity;

	if(pos.getDistance(mousePos) > 0.0f)
		gravity = (mousePos - pos).normal() * 1.0f;
	particles.trigger(pos,
					  gravity,
					  mouseCol,
					  1.0f,
					  1.0f,
					  6);
	mousePos = pos;
}

//------------------------------------------------------------------------------
void TitleScreen::mouseDown(int button, const TwoFloats& pos)
{
	if(fadeTween.getFinished())
	{
		switch(option)
		{
			case 0:
				nextState(L"GamePlayState2");
				break;
			case 1:
				globalData->setHoldMode(false);
				nextState(L"GamePlayState");
				break;
			case 2:
				globalData->setHoldMode(true);
				nextState(L"GamePlayState");
				break;
			case 3:
				fadeTween.reset(1.0f);
				break;
		}
	}
}
