//	TitleScreen.h - The title screen.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2014  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef TITLESCREEN_H_
#define TITLESCREEN_H_

#include "drawing/TextLayout.h"
#include "ParticleGenerator.h"
#include "CubicMeshDrawer.h"
#include "RainbowPalette.h"
#include "TitleHighlight.h"
#include "GameState.h"
#include "Tweener.h"

class GenericFileSounder;
class StarsGlobalData;

///	The title screen.
class TitleScreen : public GameState
{
  public:
	///	Constructor.
	TitleScreen(Application& app, SounderPtr& sound);

	///	Called frequently to update the state.
	void update();
	///	Called to draw the state.
	void draw(DrawerPtr& d, float timestepInterp);

	///	Called when a key down event is received.
	void keyDown(const std::wstring& action,
				 const std::wstring& keyPressed,
				 int keyModifier);

	///	Called when a mouse move event is received.
	void mouseMove(const TwoFloats& pos);
	///	Called when a mouse down event is received.
	void mouseDown(int button, const TwoFloats& pos);

	///	Returns the name of this state.
	const std::wstring getName() const {return L"TitleScreen";};
	///	This state's creator.
	static GameStateCreator<TitleScreen> creator;
  private:
	///	Some constants.
	enum
	{
		BackSizeX = 5,
		BackSizeY = 4,

		GridSizeX = 16,
		GridSizeY = 9
	};

	///	Top-left offset where the grid starts.
	TwoFloats gridOffset;

	///	Used to draw the background.
	CubicMeshDrawer backgroundDrawer;
	///	The grid of colours which makes up the background.
	ThreeFloats backGrid[BackSizeY][BackSizeX];
	///	Used to tween the background colours.
	std::vector<shared_ptr<Tweener<ThreeFloats> > > backTweeners;
	///	The colour palette we use to pick colours.
	RainbowPalette colourPalette;

	///	TextLayout for the title.
	TextLayout titleLayout;
	///	TextLayout for the timed mode.
	TextLayout timedLayout;
	///	TextLayout for the chaos mode.
	TextLayout chaosLayout;
	///	TextLayout for the chaos (hold) mode.
	TextLayout chaosHoldLayout;
	///	TextLayout for the quit button.
	TextLayout quitLayout;
	///	TextLayout for the expo mode text.
	TextLayout expoLayout;
	///	TextLayout for the no specials text.
	TextLayout noSpecialsLayout;

	///	Size of the timed icon.
	float timedSize;
	///	Tweener for the timed icon.
	SmoothstepTweener<float> timedTween;
	///	Size of the chaos icon.
	float chaosSize;
	///	Tweener for the chaos icon.
	SmoothstepTweener<float> chaosTween;
	///	Size of the hold icon.
	float holdSize;
	///	Tweener for the hold icon.
	SmoothstepTweener<float> holdTween;

	///	Which option we've got selected.
	int option;
	///	Highlights the current option.
	TitleHighlight highlight;

	///	The most recent mouse position.
	TwoFloats mousePos;
	///	The particle generator used indicate the mouse position.
	ParticleGenerator particles;
	///	The current particle colour.
	ThreeFloats mouseCol;
	///	The current particle colour.
	SmoothstepTweener<ThreeFloats> mouseTween;

	///	Used to fade in/out.
	float fade;
	///	Used to fade in/out.
	SmoothstepTweener<float> fadeTween;

	///	Our copy of the sounder.
	shared_ptr<GenericFileSounder> sounder;
	///	Our copy of the global data.
	shared_ptr<StarsGlobalData> globalData;
};

#endif
