//	KeyManager.cpp - Class handling the mapping of keys to actions.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "AppSettings.h"
#include "KeyManager.h"
#include "UTF8File.h"

#include <algorithm>
#include <sstream>
#include <locale>

using std::wstringstream;
using std::multimap;
using std::wstring;
using std::vector;
using std::endl;

//------------------------------------------------------------------------------
KeyManager::KeyManager():
version(1),
createFile(false)
{
	loadMappings();
}

//------------------------------------------------------------------------------
KeyManager::~KeyManager()
{
	if(createFile)
		saveMappings();
}

//------------------------------------------------------------------------------
void KeyManager::addMapping(const wstring& action,
							const wstring& key,
							bool isDefault)
{
	wstring lowerKey(key);

	//Force action lowercase.
	std::transform(lowerKey.begin(),
				   lowerKey.end(),
				   lowerKey.begin(),
				   std::bind2nd(std::ptr_fun(&std::tolower<wchar_t>), std::locale("")));

	if(isDefault)
	{
		bool dontAdd = false;
		multimap<wstring, wstring>::iterator it;

		if(!createFile)
		{
			for(it=mappings.begin();it!=mappings.end();++it)
			{
				if(it->second == action)
				{
					dontAdd = true;
					break;
				}
			}
		}
		if(!dontAdd)
			mappings.insert(make_pair(lowerKey, action));
	}
	else
		mappings.insert(make_pair(lowerKey, action));
}

//------------------------------------------------------------------------------
void KeyManager::getActions(const wstring& key, vector<wstring>& actions) const
{
	multimap<wstring, wstring>::const_iterator it;

	actions.clear();

	for(it=mappings.lower_bound(key);
		it!=mappings.upper_bound(key);
		++it)
	{
		actions.push_back(it->second);
	}
}

//------------------------------------------------------------------------------
void KeyManager::loadMappings()
{
	FilePath lastPath;
	int lastVersion(0);
	FilePath dataFolder(FilePath::getAppDataPath().getChild(AppSettings::applicationName));

	if(mappings.size() > 0)
		return;
	
	//Find most recent keyconfig<version>.txt.
	for(int i=0;i<dataFolder.getNumChildren();++i)
	{
		FilePath filePath(dataFolder.getChild(i));

		if(filePath.getName().substr(0, 9) == L"keyconfig")
		{
			int vers(-1);
			std::wstring tempstr(filePath.getName().substr(9));
			std::wstringstream tempstr2;

			tempstr2 << tempstr.substr(0, tempstr.length()-9);
			tempstr2 >> vers;

			if(vers > lastVersion)
			{
				lastVersion = vers;
				lastPath = filePath;
			}
		}
	}

	//If we've found a valid keyconfig<version>.txt, load it.
	if(lastPath.isFile() && (lastVersion >= version))
	{
		UTF8File mappingsFile(lastPath);
		wstring tempstr;
		wstring action;
		wstring key;

		version = lastVersion;

		while(!mappingsFile.readLine(tempstr))
		{
			action = tempstr.substr(0, tempstr.find_first_of(L' '));
			key = tempstr.substr(tempstr.find_first_of(L' ')+1);

			//Force key to lowercase.
			//(the awkward bind2nd() is because transform requires a unary
			// function as it's 4th parameter, and std::lower() takes 2 params
			// (the character and the locale))
			std::transform(key.begin(),
						   key.end(),
						   key.begin(), 
						   std::bind2nd(std::ptr_fun(&std::tolower<wchar_t>), std::locale("")));

			mappings.insert(make_pair(key, action));
		}
	}
	else
		createFile = true;
}

//------------------------------------------------------------------------------
void KeyManager::saveMappings()
{
	wstringstream tempstr;
	multimap<wstring, wstring>::iterator it;

	tempstr << L"keyconfig" << version << L".txt";
	FilePath filePath(FilePath::getAppDataPath().getChild(AppSettings::applicationName).getChild(tempstr.str()));
	UTF8File mappingsFile(filePath);

	for(it=mappings.begin();it!=mappings.end();++it)
	{
		tempstr.str(L"");
		tempstr.clear();

		tempstr << it->second << L" " << it->first << endl;

		mappingsFile.writeText(tempstr.str());
	}
}

//------------------------------------------------------------------------------
void KeyManager::setVersion(int val)
{
	if(version < val)
	{
		version = val;
		mappings.clear();
		createFile = true;
	}
}
