//	ScalesAndFrequencies.h - Helper stuff for working with scales.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2013  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef SCALESANDFREQUENCIES_H_
#define SCALESANDFREQUENCIES_H_

#include <vector>
#include <string>
#include <map>

///	Helper stuff for working with scales.
class ScalesAndFrequencies
{
  public:
	///	Constructor.
	ScalesAndFrequencies();

	///	Fills out the passed-in vector with a range of frequencies matching the passed-in scale.
	/*!
		\param frequencies The vector to be filled with frequencies.
		\param root The note (MIDI) at which to start the scale.
		\param octaves The number of octaves the returned frequencies should
		scale (the returned vector will contain (octaves*12)+1 entries; i.e.
		the number of octaves plus the root note a further octave up).
		\param scale The name of the scale to use. If this is empty, a random
		scale will be chosen from the available scales. If it is empty, it will
		contain the name of the chosen scale on return.
	 */
	void getFrequencies(std::vector<float>& frequencies,
						int root,
						int octaves,
						std::wstring& scale);

	///	Can be used to choose a specific set of random scales for getFrequencies().
	void chooseScales(const std::vector<std::wstring>& scaleSet);
	///	Can be used to remove a scale from the list of random scales chosen from by getFrequencies().
	void removeScale(const std::wstring& scale);
  private:
	///	Helper method. Converts an array of ones and zeros into the scales format and inserts it.
	void addScale(const std::wstring& scale, const int notes[12]);

	///	Array of the standard MIDI note frequencies.
	static float midiFreqs[128];

	///	All the available scales.
	/*!
		Stored as integer intervals (i.e. 1 == the next note in the scale
		is adjacent to the current note; 2 == the next note is two away).
	 */
	std::map<std::wstring, std::vector<int> > scales;

	///	The available random scales.
	std::vector<std::wstring> randomScales;
};

#endif
