//	StbVorbisLoader.cpp - Sound file loaders for .ogg files, using vorbisfile.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2014  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "StbVorbisLoader.h"
#include "DebugHeaders.h"
#include "UTF8File.h"

#include <cassert>

//------------------------------------------------------------------------------
StbVorbisLoader::Creator StbVorbisLoader::creator;

//------------------------------------------------------------------------------
StbVorbisLoader::StbVorbisLoader(const FilePath& path):
SoundFileLoader(path),
vorbisFile(NULL),
filePath(path),
fileLength(0),
fileSamplerate(44100.0f),
fileNumChannels(2),
valid(false),
fileBuffer(0),
fileBufferSize(16384)
{
	if(path.exists())
	{
		std::string tempstr;
		int err(VORBIS__no_error);

		UTF8File::wstringToChar(path.getPath(), tempstr);

		//Get the size of the file in bytes.
		fileSize = filePath.getFileSize();
		//Get the length of the file in samples.
		getFileLength(tempstr);

		//Setup the file stream.
		fileStream.open(tempstr, std::ios::in|std::ios::binary);
		fileBuffer = new unsigned char[fileBufferSize];

		//Read the first chunk of the file.
		fileStream.read((char *)fileBuffer, fileBufferSize);

		//Open the file.
		int numBytesConsumed(0);
		vorbisFile = stb_vorbis_open_pushdata(fileBuffer,
											  fileBufferSize,
											  &numBytesConsumed,
											  &err,
											  NULL);
		writeError(err);
		if(vorbisFile)
		{
			stb_vorbis_info info(stb_vorbis_get_info(vorbisFile));

			//Get samplerate and number of channels.
			fileSamplerate = (float)info.sample_rate;
			fileNumChannels = info.channels;

			valid = true;
		}
		else
			writeError(err);
	}
	else
		DebugStatement(L"StbVorbisLoader: There was no file at: " << path.getPath().c_str());
}

//------------------------------------------------------------------------------
StbVorbisLoader::~StbVorbisLoader()
{
	if(vorbisFile)
		stb_vorbis_close(vorbisFile);

	if(fileBuffer)
		delete [] fileBuffer;
}

//------------------------------------------------------------------------------
int StbVorbisLoader::loadChunk(float *buffer,
								int64_t& position,
								int bufferSize,
								int bufferOffset)
{
	int retval = bufferOffset;

	assert(fileNumChannels > 0);
	int numSamples = bufferSize/fileNumChannels;

	if(valid)
	{
		int arbitraryNumBytes(1024);
		uint64_t fileStreamPos;
		int numSamplesRead(0);
		float **outputBuffer;
		int nextOffset(0);
		int lastOffset(0);
		int fileOffset(0);
		int bytesUsed(0);
		int channels(2);

GonePastPosition:
		//Seek within the file to a point before the point we're looking for.
		fileStreamPos = (uint64_t)(((float)position/(float)fileLength) * (float)fileSize);
		if(fileStreamPos >= arbitraryNumBytes)
			fileStreamPos -= arbitraryNumBytes;
		else
			fileStreamPos = 0;
		fileStream.seekg(fileStreamPos);

		//Tell the decoder to resynchronise.
		stb_vorbis_flush_pushdata(vorbisFile);

		//Read a chunk of data from the file.
		fileStream.read((char *)fileBuffer, fileBufferSize);

		//Decode a frame.
		bytesUsed = stb_vorbis_decode_frame_pushdata(vorbisFile,
													 fileBuffer,
													 fileBufferSize,
													 &channels,
													 &outputBuffer,
													 &numSamplesRead);
		while(bytesUsed < 1)
		{
			bytesUsed = stb_vorbis_decode_frame_pushdata(vorbisFile,
														 fileBuffer,
														 fileBufferSize,
														 &channels,
														 &outputBuffer,
														 &numSamplesRead);
		}
		fileOffset += bytesUsed;

		//Calculate the offset of the next frame.
		nextOffset = stb_vorbis_get_sample_offset(vorbisFile);
		if((int64_t)nextOffset > position)
		{
			//Uh oh... We've gone past our desired sample position in the file.
			arbitraryNumBytes += 512;
			fileOffset -= bytesUsed;
			goto GonePastPosition; //Ugh, goto...
		}

		//Search for the frame containing our desired sample position.
		while(position >= nextOffset)
		{
			//Decode the next frame.
			lastOffset = nextOffset;
			bytesUsed = stb_vorbis_decode_frame_pushdata(vorbisFile,
														 fileBuffer + fileOffset,
														 fileBufferSize,
														 &channels,
														 &outputBuffer,
														 &numSamplesRead);
			fileOffset += bytesUsed;

			//The decoder wants more data.
			//(potential runaway memory usage, yay!)
			if(bytesUsed == 0)
			{
				fileBufferSize += 1024;
				delete [] fileBuffer;
				fileBuffer = new unsigned char[fileBufferSize];

				//Re-read the file data.
				fileStream.seekg(fileStreamPos);
				fileStream.read((char *)fileBuffer, fileBufferSize);

				stb_vorbis_flush_pushdata(vorbisFile);
				writeError(stb_vorbis_get_error(vorbisFile));
			}

			//Calculate the offset of the next frame.
			nextOffset = stb_vorbis_get_sample_offset(vorbisFile);
			if((position >= lastOffset) && (position < nextOffset) && (numSamplesRead > 0))
				break;

			//If necessary, read some more data from the file.
			if(fileOffset >= fileBufferSize)
			{
				fileStream.read((char *)fileBuffer, fileBufferSize);
				fileOffset = 0;
			}
		}
		
		while(retval < numSamples)
		{
			//Copy the contents of outputBuffer (from our desired position) to
			//buffer.
			int outputBufferOffset((int)(position-lastOffset));

			for(int i=0;i<(nextOffset-position);++i)
			{
				for(int j=0;j<channels;++j)
				{
					buffer[retval] = outputBuffer[j][outputBufferOffset+i];

					++retval;
					if(retval >= numSamples)
						break;
				}
				if(retval >= numSamples)
					break;
			}

			if(retval < numSamples)
			{
				int bytesUsed(0);

NoSamplesRead:
				//Decode the next frame.
				lastOffset = nextOffset;
				bytesUsed = stb_vorbis_decode_frame_pushdata(vorbisFile,
															 fileBuffer + fileOffset,
															 fileBufferSize,
															 &channels,
															 &outputBuffer,
															 &numSamplesRead);
				fileOffset += bytesUsed;

				//Calculate the offset of the next frame.
				nextOffset = stb_vorbis_get_sample_offset(vorbisFile);

				//If necessary, read some more data from the file.
				if(fileOffset >= fileBufferSize)
				{
					fileStream.read((char *)fileBuffer, fileBufferSize);
					fileOffset = 0;
				}

				if(!numSamplesRead)
					goto NoSamplesRead;
			}
		}
	}

	return (retval-bufferOffset);
}

//------------------------------------------------------------------------------
void StbVorbisLoader::getFileLength(const std::string& narrowPath)
{
	int err(VORBIS__no_error);
	stb_vorbis *tempFile(stb_vorbis_open_filename((char *)narrowPath.c_str(),
												  &err,
												  NULL));
	if(tempFile)
	{
		//Get the length of the file.
		fileLength = stb_vorbis_stream_length_in_samples(tempFile);

		stb_vorbis_close(tempFile);
	}
}

//------------------------------------------------------------------------------
void StbVorbisLoader::writeError(int err)
{
	switch(err)
	{
		case VORBIS__no_error:
			DebugStatement(L"StbVorbisLoader: no error");
			break;
		case VORBIS_need_more_data:
			DebugStatement(L"StbVorbisLoader: Need more data (not a real error).");
			break;
		case VORBIS_invalid_api_mixing:
			DebugStatement(L"StbVorbisLoader: Invalid API mixing (can't mix modes).");
			break;
		case VORBIS_outofmem:
			DebugStatement(L"StbVorbisLoader: Out of memory");
			break;
		case VORBIS_feature_not_supported:
			DebugStatement(L"StbVorbisLoader: Feature not supported (uses floor 0).");
			break;
		case VORBIS_too_many_channels:
			DebugStatement(L"StbVorbisLoader: Too many channels (STB_VORBIS_MAX_CHANNELS).");
			break;
		case VORBIS_file_open_failure:
			DebugStatement(L"StbVorbisLoader: File open failure (fopen() failed).");
			break;
		case VORBIS_seek_without_length:          // can't seek in unknown-length file:
			DebugStatement(L"StbVorbisLoader: Unknown length of file (can't seek in a file with unknown length).");
			break;
		case VORBIS_unexpected_eof:            // file is truncated?:
			DebugStatement(L"StbVorbisLoader: Unexpected EOF (file is truncated?)");
			break;
		case VORBIS_seek_invalid:                 // seek past EOF:
			DebugStatement(L"StbVorbisLoader: Seek invalid (passed EOF).");
			break;
		case VORBIS_invalid_setup:
			DebugStatement(L"StbVorbisLoader: Decoding error (invalid setup).");
			break;
		case VORBIS_invalid_stream:
			DebugStatement(L"StbVorbisLoader: Decoding error (invalid stream).");
			break;
		case VORBIS_missing_capture_pattern:
			DebugStatement(L"StbVorbisLoader: Decoding error (Missing capture pattern).");
			break;
		case VORBIS_invalid_stream_structure_version:
			DebugStatement(L"StbVorbisLoader: Decoding error (invalid stream structure version).");
			break;
		case VORBIS_continued_packet_flag_invalid:
			DebugStatement(L"StbVorbisLoader: Decoding error (continued packet flag invalid).");
			break;
		case VORBIS_incorrect_stream_serial_number:
			DebugStatement(L"StbVorbisLoader: Decoding error (incorrect stream serial number).");
			break;
		case VORBIS_invalid_first_page:
			DebugStatement(L"StbVorbisLoader: Decoding error (invalid first page).");
			break;
		case VORBIS_bad_packet_type:
			DebugStatement(L"StbVorbisLoader: Decoding error (bad packet type).");
			break;
		case VORBIS_cant_find_last_page:
			DebugStatement(L"StbVorbisLoader: Decoding error (can't find last page).");
			break;
		case VORBIS_seek_failed:
			DebugStatement(L"StbVorbisLoader: Decoding error (seek failed).");
	}
}
