//	StbVorbisLoader.h - Sound file loaders for .ogg files, using stb_vorbis.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2014  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef STBVORBISLOADER_H_
#define STBVORBISLOADER_H_

#include "../SoundFileLoader.h"

#include "stb_vorbis.h"

#include <fstream>

///	Sound file loaders for .ogg files, using vorbisfile.
class StbVorbisLoader : public SoundFileLoader
{
  public:
	///	Constructor.
	StbVorbisLoader(const FilePath& path);
	///	Destructor.
	~StbVorbisLoader();

	///	Loads a chunk of the file into the passed-in buffer.
	int loadChunk(float *buffer,
				  int64_t& position,
				  int bufferSize,
				  int bufferOffset = 0);

	///	Returns the length of the sound file in samples.
	int64_t getFileLength() const {return fileLength;};
	///	Returns the sound file's samplerate.
	float getSamplerate() const {return fileSamplerate;};
	///	Returns the number of channels in the file.
	int getNumChannels() const {return fileNumChannels;};

	///	Returns the name of the sound file.
	const std::wstring getName() const {return filePath.getName();};

	///	Used to create StbVorbisLoaders
	class Creator : public SoundFileLoader::Creator
	{
	  public:
		///	Constructor.
		Creator() {};
		///	Destructor.
		~Creator() {};

		///	Returns a new instance of a StbVorbisLoader.
		SoundFileLoaderPtr createLoader(const FilePath& path) const
		{
			return SoundFileLoaderPtr(new StbVorbisLoader(path));
		};
	};

	///	The sole instance of Creator for this class.
	static Creator creator;
  private:
	///	Hide the copy constructor (rule of 3).
	StbVorbisLoader(const StbVorbisLoader& other):
	SoundFileLoader(other.filePath)
	{
	
	};
	///	Hide the assignment operator (rule of 3).
	StbVorbisLoader& operator=(const StbVorbisLoader& other) {return *this;};

	///	Helper method. Gets the length of the file in samples.
	void getFileLength(const std::string& narrowPath);
	///	Helper method. Writes an error to the debug log.
	void writeError(int err);

	///	The sound file's path.
	FilePath filePath;
	///	Our file stream.
	std::ifstream fileStream;
	///	Handle to the sound file.
	stb_vorbis *vorbisFile;

	///	The size in bytes of the sound file.
	uint64_t fileSize;
	///	The length of the sound file.
	int64_t fileLength;
	///	The samplerate of the sound file.
	float fileSamplerate;
	///	The number of channels in the file.
	int fileNumChannels;

	///	True if vorbisFile was setup correctly in the constructor.
	bool valid;

	///	The buffer we read the file data into.
	unsigned char *fileBuffer;
	///	The current size of the fileBuffer.
	int fileBufferSize;
};

#endif
